
import json 

def parse_json2str(json_file): 
    str2json = {}
    for key in json_file: 
        str2json[key] = str(json_file[key])
    return str2json

def parse_str2json(json_file): 
    unstr2json = {}
    for key in json_file: 
        string = json_file[key]
        if string.startswith(("{", "[")): 
            unstr2json[key] = eval(json_file[key])
        else: 
            unstr2json[key] = json_file[key]
    return unstr2json

def parse_json4kaapana(json_file): 
    myDict = {}
    myDict["predicted_bodypart_string"] = json_file["body part examined tag"]
    myDict["prediction_parameters_string"] = json_file
    
    return parse_json2str(myDict)

def test_parser(): 
    myDict = {"a": [1, 2, 3, 4], 
              "b": "Hi", 
              "c": {"l1": 1, "l2": 2, "l3": 3}, 
              "d": {"l1": {"a": 0, "b": 0}}}
    
    j1 = parse_json2str(myDict)
    print(j1)

    j2 = parse_str2json(j1)
    print(j2)
    print(j2==myDict)
  
if __name__ == "__main__": 
    test_parser()



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys, json, pickle

sys.path.append("../../")

from dataclasses import dataclass, field
from torchvision import transforms
import albumentations as A

from bpreg.settings.settings import *
from bpreg.dataset.custom_transformations import *


@dataclass
class ModelSettings:
    """Create settings file for training a body part regression model."""

    df_data_source_path: str = ""
    data_path: str = ""
    landmark_path: str = ""
    save_dir: str = ""
    batch_size: int = 64
    effective_batch_size: int = 64
    equidistance_range: list = field(default_factory=lambda: [5, 100])
    num_slices: int = 4
    epochs: int = 480
    alpha_h: float = 1
    beta_h: float = 0.01
    loss_order: str = "h"
    lambda_: float = 0
    alpha: float = 0
    lr: float = 1e-4
    shuffle_train_dataloader: bool = True
    random_seed: int = 0
    deterministic: bool = True
    save_model: bool = True
    base_model: str = "vgg"
    transform_params: dict = field(default_factory=lambda: TRANSFORM_STANDARD_PARAMS)
    name: str = "default.p"
    model_name: str = "standard"
    model: str = ""

    def __post_init__(self):
        self.json_dict = self.__dict__.copy()

        self.filepath = self.save_dir + self.name
        self.custom_transform = self.get_custom_transform()
        self.albumentation_transform = self.get_albumentation_transform()

    def get_custom_transform(self):
        custom_transforms = []
        for transform in [GaussNoise, ShiftHU, ScaleHU, AddFrame]:
            transform_name = transform.__name__
            if not transform_name in self.transform_params:
                continue
            params = self.transform_params[transform_name]
            custom_transforms.append(transform(**params))
        return transforms.Compose(custom_transforms)

    def get_albumentation_transform(self):
        albumetnation_transforms = []
        for transform in [A.Flip, A.Transpose, A.ShiftScaleRotate, A.GaussianBlur]:
            transform_name = transform.__name__
            if not transform_name in self.transform_params:
                continue
            params = self.transform_params[transform_name]
            albumetnation_transforms.append(transform(**params))
        return A.Compose(albumetnation_transforms)

    def save(self, save_path=""):
        for key in [
            "albumentation_transform",
            "custom_transform",
            "filepath",
            "json_dict",
        ]:
            if key in self.json_dict:
                del self.json_dict[key]

        if len(save_path) == "":
            save_path = self.filepath
        with open(save_path, "w") as f:
            json.dump(self.json_dict, f, indent=4)

    def load(self, path=""):
        with open(path, "r") as f:
            json_dict = json.load(f)

        self.__init__(**json_dict)
        self.__post_init__()

    def load_pickle(self, path):
        """Provide backward compatibility with outdated pickle files."""

        delete_keys = [
            "custom_transform",
            "albumentation_transform",
            "custom_transform_params",
            "albumentation_transform_params",
            " ",
            "   ",
            "     ",
            "    ",
            "description",
            "accumulate_grad_batches",
            "lambda",
            "test_loss",
            "test_loss_order",
            "test_loss_dist",
            "test_loss_l2",
            "test_landmark_metric_mean",
            "test_landmark_metric_var",
            "pearson-correlation",
            "validation loss",
            "landmark metric",
            "pre-name",
        ]

        with open(path, "rb") as f:
            config = pickle.load(f)

        custom_transform = config["custom_transform"]
        albumentation_transform = config["albumentation_transform"]
        config["lambda_"] = config["lambda"]

        # delete keys which this class does not need
        for key in delete_keys:
            if key in config.keys():
                del config[key]

        # overwrite parameters
        self.__init__(**config)
        self.custom_transform = custom_transform
        self.albumentation_transform = albumentation_transform

        # get transform parameters
        params = self.transforms_to_dict(
            self.custom_transform.__dict__["transforms"],
            ["square_frame", "circle_frame"],
        )
        params_albumentation = self.transforms_to_dict(
            self.albumentation_transform,
            [
                "deterministic",
                "save_key",
                "replay_mode",
                "mask_value",
                "applied_in_replay",
                "params",
            ],
            albumentation=True,
        )
        params.update(params_albumentation)

        # if "ShiftHU" in params.keys():
        #   params["ShiftHU"]["limit"] = params["ShiftHU"]["shift_limit"]
        #   del params["ShiftHU"]["shift_limit"]

        if "AddFrame" in params.keys():
            params["AddFrame"]["dimension"] = params["AddFrame"]["d"]
            del params["AddFrame"]["d"]

        self.transform_params = params
        self.json_dict = self.__dict__.copy()

    def transforms_to_dict(self, transform_list, delete_keys, albumentation=False):
        params = {}
        for transform in transform_list:
            name = type(transform).__name__

            transform_params = transform.__dict__
            if albumentation:
                transform_params = transform.get_base_init_args()
                transform_params.update(transform.get_transform_init_args())

            params[name] = transform_params

            # filter private keys
            params[name] = {
                key: params[name][key]
                for key in params[name].keys()
                if not key.startswith("_")
            }

            for key in delete_keys:
                if key in params[name].keys():
                    del params[name][key]

        return params

    def __str__(self):
        start_line = "\nBODY PART REGRESSION MODEL SETTINGS\n"
        line = "*******************************************************\n"
        part1_keys = [
            "model_name",
            "name",
            "df_data_source_path",
            "data_path",
            "landmark_path",
            "save_dir",
            "shuffle_train_dataloader",
            "random_seed",
            "deterministic",
            "save_model",
            "base_model",
        ]
        part2_keys = [
            "batch_size",
            "effective_batch_size",
            "equidistance_range",
            "num_slices",
        ]
        part3_keys = [
            "alpha_h",
            "beta_h",
            "loss_order",
            "lambda_",
            "alpha",
            "lr",
            "epochs",
        ]

        my_string = start_line + line
        for part in [part1_keys, part2_keys, part3_keys]:
            for key in part:
                my_string += f"{key:<28}:\t{self.json_dict[key]}\n"
            my_string += line
        my_string += "\n"

        return my_string



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import datetime
import random, sys
import torch
import cv2
import albumentations as A
from scipy.stats import pearsonr
import pytorch_lightning as pl
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torchvision.models as models

cv2.setNumThreads(1)

sys.path.append("../../")
from bpreg.evaluation.landmark_mse import LMSE
from bpreg.network_architecture.loss_functions import *


class BodyPartRegressionBase(pl.LightningModule):
    def __init__(
        self,
        lr: float = 1e-4,
        lambda_: float = 0,
        alpha: float = 0,
        pretrained: bool = False,
        delta_z_max: float = np.inf,
        loss_order: str = "h",
        beta_h: float = 0.025,
        alpha_h: float = 0.5,
        weight_decay: float = 0,
    ):

        super().__init__()
        self.lr = lr
        self.alpha_h = alpha_h
        self.beta_h = beta_h
        self.alpha = alpha
        self.weight_decay = weight_decay
        self.loss_order_name = loss_order
        self.delta_z_max = delta_z_max
        self.l1loss = torch.nn.SmoothL1Loss(reduction="mean")
        self.pretrained = pretrained
        self.lambda_ = lambda_
        self.val_landmark_metric = []
        self.val_loss = []
        self.save_hyperparameters()        
        self.lmse = LMSE()

        if loss_order == "h":
            self.loss_order = order_loss_h(alpha=self.alpha_h, beta=self.beta_h)
        elif loss_order == "c":
            self.loss_order = order_loss_c()

        elif loss_order == "":
            self.loss_order = no_order_loss()
        else:
            raise ValueError(f"Unknown loss parameter {loss_order}")

    def base_step(self, batch, batch_idx):
        x, slice_indices, z = batch
        x, batch_size, num_slices = self.to1channel(x)
        y_hat = self(x)
        y_hat = self.tonchannel(y_hat, batch_size, num_slices)
        loss, loss_order, loss_dist, loss_l2 = self.loss(y_hat, slice_indices, z)
        return loss, loss_order, loss_dist, loss_l2

    def training_step(self, batch, batch_idx):
        loss, loss_order, loss_dist, loss_l2 = self.base_step(batch, batch_idx)
        self.log("train_loss", loss)
        self.log("train_loss_order", loss_order)
        self.log("train_loss_dist", loss_dist)
        self.log("train_loss_l2", loss_l2)
        return loss

    def to1channel(self, x):
        batch_size = x.shape[0]
        num_slices = x.shape[1]
        x = x.reshape(batch_size * num_slices, 1, x.shape[2], x.shape[3])
        return x, batch_size, num_slices

    def tonchannel(self, x, batch_size, num_slices):
        x = x.reshape(batch_size, num_slices)
        return x

    def validation_epoch_end(self, validation_step_outputs):
        val_dataloader = self.val_dataloader()
        train_dataloader = self.train_dataloader()

        lmse, lmse_std, d = self.lmse.from_dataset(
            self, val_dataloader.dataset, train_dataloader.dataset
        )

        self.log("lmse", lmse)
        self.log("lmse_std", lmse_std)
        self.log("d", d)

    def validation_step(self, batch, batch_idx):
        loss, loss_order, loss_dist, loss_l2 = self.base_step(batch, batch_idx)
        self.log("val_loss", loss)
        self.log("val_loss_order", loss_order)
        self.log("val_loss_dist", loss_dist)
        self.log("val_loss_l2", loss_l2)

    def test_step(self, batch, batch_idx):
        test_dataloader = self.test_dataloader()
        train_dataloader = self.train_dataloader()

        lmse, lmse_std, d = self.lmse.from_dataset(
            self, test_dataloader.dataset, train_dataloader.dataset
        )

        self.log("lmse", lmse)
        self.log("lmse_std", lmse_std)
        self.log("d", d)

    def loss(self, scores_pred, slice_indices, z):
        l2_norm = 0
        ldist_reg = 0
        loss_order = self.loss_order(scores_pred, z)
        if self.lambda_ > 0:
            l2_norm = self.lambda_ * torch.mean(scores_pred ** 2)
        if self.alpha > 0:
            ldist_reg = self.alpha * self.loss_dist(scores_pred, z)
        loss = loss_order + l2_norm + ldist_reg
        return loss, loss_order, ldist_reg, l2_norm

    def loss_dist(self, scores_pred, z):
        mask = torch.where(z > self.delta_z_max, 0, 1)
        scores_diff = (scores_pred[:, 1:] - scores_pred[:, :-1]) * mask
        loss = self.l1loss(scores_diff[:, 1:], scores_diff[:, :-1])
        return loss

    def configure_optimizers(self):
        optimizer = optim.Adam(
            self.parameters(), lr=self.lr, weight_decay=self.weight_decay
        )
        return optimizer

    def compute_slice_score_matrix(self, dataset, inference_device="cuda"):
        with torch.no_grad():
            self.eval()
            self.to(inference_device)
            slice_score_matrix = np.full(dataset.landmark_matrix.shape, np.nan)

            for i, slices, defined_landmarks in zip(
                np.arange(0, slice_score_matrix.shape[0]),
                dataset.landmark_slices_per_volume,
                dataset.defined_landmarks_per_volume,
            ):
                scores = self(
                    torch.tensor(slices[:, np.newaxis, :, :]).to(inference_device)
                )
                slice_score_matrix[i, defined_landmarks] = (
                    scores[:, 0].cpu().detach().numpy()
                )
        return slice_score_matrix

    def predict_tensor(self, tensor, n_splits=200, inference_device="cuda"):
        scores = []
        n = tensor.shape[0]
        slice_splits = list(np.arange(0, n, n_splits))
        slice_splits.append(n)

        with torch.no_grad():
            self.eval()
            self.to(inference_device)
            for i in range(len(slice_splits) - 1):
                min_index = slice_splits[i]
                max_index = slice_splits[i + 1]
                score = self(tensor[min_index:max_index, :, :, :].to(inference_device))
                scores += [s.item() for s in score]

        scores = np.array(scores)
        return scores

    def predict_npy_array(self, x, n_splits=200, inference_device="cuda"):
        x_tensor = torch.tensor(x[:, np.newaxis, :, :]).to(inference_device)
        scores = self.predict_tensor(
            x_tensor, inference_device=inference_device, n_splits=n_splits
        )
        return scores



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys, os, argparse

sys.path.append("../")
from bpreg.settings import ModelSettings

data_path = {
    "local": "/home/AD/s429r/Documents/Data/DataSet/",
    "cluster": "/gpu/data/OE0441/s429r/",
}

save_path = {
    "local": "/home/AD/s429r/Documents/Data/Results/body-part-regression-models/",
    "cluster": "/gpu/checkpoints/OE0441/s429r/results/bodypartregression/",
}


def base_config(mode: str = "cluster"):
    params = {
        "df_data_source_path": data_path[mode]
        + "MetaData/meta-data-public-dataset-npy-arrays-3.5mm-windowing-sigma.xlsx",
        "data_path": data_path[mode] + "Arrays-3.5mm-sigma-01/",
        "landmark_path": data_path[mode] + "MetaData/landmarks-meta-data-v2.xlsx",
        "save_dir": save_path[mode],
    }
    return params


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--mode", default="cluster")

    value = parser.parse_args()
    mode = value.mode

    experiments = {0: {"model_name": "standard", "name": "standard-2.p"}}
    base_path = "../src/configs/" + mode + "/"

    for idx, data in experiments.items():
        params = base_config(mode=mode)
        params.update(data)
        config = ModelSettings(**params)
        print(config)
        save_path_dir = base_path + config.model_name + "/"
        if not os.path.exists(save_path_dir):
            os.mkdir(save_path_dir)
        print(f"SAVE CONFIG: {save_path_dir + config.name}")
        config.save(save_path=save_path_dir + config.name)



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from scipy import interpolate
import matplotlib.pyplot as plt
import numpy as np

from bpreg.score_processing import Scores
from bpreg.utils.linear_transformations import *
from bpreg.dataset.base_dataset import BaseDataset



class ValidationVolume:
    def __init__(
        self,
        inference_model,
        val_dataset: BaseDataset,
        idx: int,
        expected_scores: np.array,
        fontsize: float = 22,
    ):
        self.landmark_idx = idx
        self.data_idx = val_dataset.landmark_ids[idx]

        self.landmarks = val_dataset.landmark_matrix[self.landmark_idx, :]
        self.landmark_names = val_dataset.landmark_names

        self.nonempty = np.where(~np.isnan(self.landmarks))[0]
        self.filename = val_dataset.filenames[self.data_idx]
        self.expected_scores = expected_scores
        self.X = val_dataset.get_full_volume(self.data_idx)
        self.z = val_dataset.z_spacings[self.data_idx]
        self.scores = inference_model.predict_npy_array(self.X)
        self.scores = Scores(
            self.scores,
            self.z,
            transform_min=expected_scores[0],
            transform_max=expected_scores[-1],
        )
        self.interpolated_x, self.interpolated_scores = self.get_interpolated_scores(
            self.landmarks, self.expected_scores
        )
        self.z_array = np.arange(0, len(self.X)) * self.z
        self.fontsize = fontsize
        self.figsize = (14, 8)
        self.markerstyles = np.array(
            [".", "v", "*", "p", "D", "X", "h", "P", "+", "^", "x", "d"]
        )
        self.colors = np.array(
            [
                "firebrick",
                "sienna",
                "tan",
                "olivedrab",
                "lightseagreen",
                "paleturquoise",
                "royalblue",
                "navy",
                "mediumpurple",
                "mediumvioletred",
                "crimson",
                "saddlebrown",
            ]
        )

    def plot_scores(self, set_figsize=False, legend=None):
        if set_figsize:
            plt.figure(figsize=self.figsize)
        label = None

        # plot scores
        plt.plot(
            self.z_array,
            self.scores.original_transformed_values,
            color="black",
            linewidth=1,
        )

        # plot landmarks
        for name, landmark, color, score in zip(
            self.landmark_names[self.nonempty],
            self.landmarks[self.nonempty],
            self.colors[self.nonempty],
            self.scores.original_transformed_values[
                self.landmarks[self.nonempty].astype(int)
            ],
        ):
            if legend:
                label = name
            plt.plot(
                landmark * self.z,
                score,
                linestyle="",
                marker=".",
                markersize=15,
                color=color,
                label=label,
            )

        plt.xlabel("z [mm]", fontsize=self.fontsize)
        plt.ylabel("Slice Score", fontsize=self.fontsize)
        plt.xticks(fontsize=self.fontsize - 2)
        plt.yticks(fontsize=self.fontsize - 2)
        if legend:
            plt.legend(loc=0)

    def transform(self, x):
        return linear_transform(
            x,
            scale=100,
            min_value=self.expected_scores[0],
            max_value=self.expected_scores[-1],
        )

    def plot_expected_scores(
        self, ax=False, title=None, legend=True, ylim=(-10, 80), ylabel=True
    ):
        if not ax:
            fig, ax = plt.subplots(figsize=(self.figsize))
        ax.plot(
            self.z_array,
            self.scores.original_transformed_values,
            linewidth=2,
            label="predicted scores",
        )

        ax.plot(
            self.landmark_positions * self.z,
            self.landmark_scores,
            linestyle="",
            marker="+",
            markersize=15,
            color="#ff7f0e",
        )
        ax.plot(
            self.interpolated_x,
            self.interpolated_scores,
            linestyle="--",
            color="#ff7f0e",
            linewidth=2,
            label="expected scores",
        )
        if legend:
            ax.legend(fontsize=self.fontsize)
        ax.tick_params(labelsize=self.fontsize - 2)
        ax.tick_params(labelsize=self.fontsize - 2)
        ax.set_xlabel("z [mm]", fontsize=self.fontsize)
        if ylabel:
            ax.set_ylabel("Slice scores", fontsize=self.fontsize)
        ax.set_title(title, fontsize=self.fontsize)
        ax.set_ylim(ylim)

    def get_interpolated_scores(self, landmarks, expected_scores):
        nonempty = np.where(~np.isnan(landmarks))
        x = np.arange(np.nanmin(landmarks), np.nanmax(landmarks) + 1)

        self.landmark_positions = landmarks[nonempty]
        self.landmark_scores = self.transform(expected_scores[nonempty])
        self.defined_landmarks = self.landmark_names[nonempty]
        func = interpolate.interp1d(
            self.landmark_positions, self.landmark_scores, kind="linear"
        )
        interpolated_scores = func(x)

        return x * self.z, interpolated_scores



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os, sys, json
import numpy as np
import matplotlib.pyplot as plt
import torch

sys.path.append("../../")
from bpreg.utils.training_utils import get_dataframe, get_datasets
from bpreg.evaluation.landmark_mse import LMSE
from bpreg.evaluation.visualization import Visualization
from bpreg.inference.inference_model import InferenceModel
from bpreg.score_processing import LandmarkScoreBundle
from bpreg.settings import *


class Evaluation(Visualization):
    """evaluate body part regression model

    Args:
        base_filepath (str): path of trained model
        df_data_source_path (str, optional): Path to dataframe which defines train/val/test split. Defaults to DF_DATA_SOURCE_PATH.
        landmark_path (str, optional): Path to dataframe which saves the annotated landmarks. Defaults to LANDMARK_PATH.
        data_path (str, optional): Path where the .npy files for training are saved. Defaults to DATA_PATH.
        device (str, optional): device, where to save the model. Defaults to "cuda".
        landmark_start (str, optional): start landmark gets mapped through transformation to zero. Defaults to "pelvis_start".
        landmark_end (str, optional): end_landmark gets mapped through transformation to 100. Defaults to "eyes_end".
    """

    def __init__(
        self,
        base_filepath: str,
        df_data_source_path: str = DF_DATA_SOURCE_PATH,
        landmark_path: str = LANDMARK_PATH,
        data_path: str = DATA_PATH,
        device: str = "cuda",
        landmark_start: str = "pelvis_start",
        landmark_end: str = "eyes_end",
    ):

        Visualization.__init__(self)
        self.device = device
        self.base_filepath = base_filepath
        self.inference_model = InferenceModel(base_filepath)
        self.normalizedMSE = LMSE()
        self.landmark_score_bundle = LandmarkScoreBundle(
            data_path,
            landmark_path,
            self.inference_model.model,
            landmark_start=landmark_start,
            landmark_end=landmark_end,
        )

        # setup data
        self.df_data_source_path = df_data_source_path
        self.landmark_path = landmark_path  
        self.data_path = data_path
        self._setup_data()

        self.mse, self.mse_std = self.landmark_score_bundle.nMSE(
            target="validation", reference="train"
        )
        self.acc5 = self.landmark_score_bundle.accuracy(
            self.val_dataset, reference="train", class2landmark=CLASS_TO_LANDMARK_5
        )

    def _setup_data(self):
        path = self.base_filepath + "config.json"

        self.config = ModelSettings()
        self.config.load(path)

        config = self.config
        config.num_slices = 8
        config.batch_size = 32
        config.shuffle_train_dataloader = False

        if len(self.df_data_source_path) > 0:
            config.df_data_source_path = self.df_data_source_path
        if len(self.landmark_path) > 0:
            config.landmark_path = self.landmark_path
        if len(self.data_path) > 0:
            config.data_path = self.data_path

        df_data = get_dataframe(config)

        self.train_dataset, self.val_dataset, self.test_dataset = get_datasets(
            config, df_data
        )

        self.train_dataloader = torch.utils.data.DataLoader(
            self.train_dataset,
            batch_size=config.batch_size,
            num_workers=20,
            shuffle=config.shuffle_train_dataloader,
        )

        self.val_dataloader = torch.utils.data.DataLoader(
            self.val_dataset, batch_size=config.batch_size, num_workers=20
        )

        self.test_dataloader = torch.utils.data.DataLoader(
            self.test_dataset, batch_size=config.batch_size, num_workers=20
        )

    def print_summary(self):
        print("Model summary\n*******************************")

        print(
            f"\nLandmark Mean Square Error:\t{self.mse:<1.3f} +- {self.mse_std:<1.3f}"
        )
        print(f"Accuracy (5 classes) :        \t{self.acc5*100:<1.2f}%")
        print("\nLook-up table for training data \n*******************************")
        self.landmark_score_bundle.dict["train"].print_lookuptable()

    def plot_landmarks(self, colors=[], alpha=0.7, target="validation"):
        if target == "validation":
            score_matrix = self.landmark_score_bundle.dict[
                "validation"
            ].score_matrix_transformed
        if target == "test":
            score_matrix = self.landmark_score_bundle.dict[
                "test"
            ].score_matrix_transformed
        expected_scores = self.landmark_score_bundle.dict[
            "train"
        ].expected_scores_transformed

        landmark_names = self.landmark_score_bundle.dict["validation"].landmark_names
        super(Evaluation, self).plot_landmarks(
            score_matrix,
            expected_scores=expected_scores,
            landmark_names=landmark_names,
            alpha=alpha,
            colors=colors,
        )

    def plot_slices2scores(
        self, max_cols=4, nearby_values=[0, 25, 50, 75, 100], save_path="", fontsize=16
    ):
        _, ax = plt.subplots(len(nearby_values), max_cols, figsize=(14, 16))

        for row, nearby_value in enumerate(nearby_values):
            col = 0
            slice_indices = []
            while col < max_cols:
                idx = np.random.randint(0, len(self.test_dataset))
                if idx in slice_indices:
                    continue
                slice_indices.append(idx)

                filepaths = self.test_dataset.filepaths
                zspacing = self.test_dataset.z_spacings[idx]
                X = np.load(filepaths[idx]).transpose(2, 0, 1)
                scores = self.inference_model.predict_npy_array(X)
                myScores = self.inference_model.parse_scores(scores, zspacing)
                slice_index = np.argmin(
                    np.abs(myScores.original_transformed_values - nearby_value)
                )
                slice_score = myScores.original_transformed_values[slice_index]
                if np.abs(slice_score - nearby_value) > 1:
                    continue

                ax[row, col].imshow(X[slice_index, :, :], cmap="gray")
                ax[row, col].set_title(np.round(slice_score, 2), fontsize=fontsize)
                ax[row, col].set_yticklabels([])
                ax[row, col].set_xticklabels([])
                ax[row, col].set_yticks([])
                ax[row, col].set_xticks([])
                col += 1
        plt.tight_layout()
        if len(save_path) > 0:
            plt.savefig(save_path)


if __name__ == "__main__":
    base_dir = "/home/AD/s429r/Documents/Code/bodypartregression/src/models/loh-ldist-l2/sigma-dataset-v11-v2/"
    modelEval = Evaluation(base_dir)



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch


class order_loss_h:
    """
    Heuristic order loss
    """

    def __init__(self, alpha: float, beta: float):
        self.alpha = alpha
        self.beta = beta

    def __call__(self, scores_pred, z):
        scores_diff = scores_pred[:, 1:] - scores_pred[:, :-1]
        p_pred = torch.sigmoid(self.alpha * scores_diff)
        p_obs = torch.sigmoid(self.beta * z)
        loss = torch.mean((p_obs - p_pred) ** 2)
        return loss


class order_loss_c:
    """
    Classification order loss
    """

    def __init__(self):
        pass

    def __call__(self, scores_pred, _):
        scores_diff = scores_pred[:, 1:] - scores_pred[:, :-1]
        sigmoid = torch.sigmoid(scores_diff)

        # remove zeros --> they can lead to nan values
        sigmoid = sigmoid[sigmoid > 0]
        loss = -torch.mean(torch.log(sigmoid))
        return loss


class order_loss_c_plain:
    """
    Classification order loss
    """

    def __init__(self):
        pass

    def __call__(self, scores_pred, _):
        scores_diff = scores_pred[:, 1:] - scores_pred[:, :-1]
        sigmoid = torch.sigmoid(scores_diff)
        loss = -torch.mean(torch.log(sigmoid))
        return loss


class no_order_loss:
    def __init__(self):
        pass

    def __call__(self, scores, _):
        return 0.0



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from bpreg.inference.inference_model import InferenceModel
from bpreg.dataset.bpr_dataset import BPRDataset
from bpreg.settings.settings import DATA_PATH, DF_DATA_SOURCE_PATH, LANDMARK_PATH
import numpy as np
import json

from tqdm import tqdm
from dataclasses import dataclass, field
from bpreg.score_processing import Scores
from bpreg.evaluation.evaluation import Evaluation
from torch.utils.data import Dataset


def postprocess_model_for_inference(
    model_path: str,
    df_data_source_path: str = DF_DATA_SOURCE_PATH,
    landmark_path: str = LANDMARK_PATH,
    data_path: str = DATA_PATH,
    upper_tangential_slope_quantile: float = 0.995,
    lower_tangential_slope_quantile: float = 0.005,
    transform_min_landmark: str = "pelvis_start",
    transform_max_landmark: str = "eyes_end",
    skip_slopes: bool = False,
):
    """Create inference-settings.json for model.
    The inference-settings.json file saves all relevant information which is needed for using the model during test time.

    Args:
        model_path (str): path of directory, where model.pt and config.json file is inside
        df_data_source_path (str, optional): path of the excel file which saves the train/val/test split
        landmark_path (str, optional): path of the excel file which saves the annotations for the train/val and test data
        upper_tangential_slope_quantile (float, optional): quantile of upper boundary for valid tangential slopes
        lower_tangential_slope_quantile (float, optional):  quantile of lower boundary for valid tangential slopes
        transform_min_landmark (str, optional): landmark which should get mapped to 0 after transformation
        transform_max_landmark (str, optional):  landmark which should get mapped to 100 after transformation
    """

    # get model evaluation
    print("Initialize model.")
    modelEval = Evaluation(
        model_path,
        df_data_source_path=df_data_source_path,
        landmark_path=landmark_path,
        data_path=data_path,
        landmark_start=transform_min_landmark,
        landmark_end=transform_max_landmark,
    )

    # get lookuptables
    print("Load lookuptables")
    lookuptable_train_original = modelEval.landmark_score_bundle.dict[
        "train"
    ].lookuptable
    lookuptable_train_transformed = modelEval.landmark_score_bundle.dict[
        "train"
    ].transformed_lookuptable

    lookuptable_train = {
        "original": lookuptable_train_original,
        "transformed": lookuptable_train_transformed,
    }

    lookuptable_train_val_original = modelEval.landmark_score_bundle.dict[
        "train+val-all-landmarks"
    ].lookuptable
    lookuptable_train_val_transformed = modelEval.landmark_score_bundle.dict[
        "train+val-all-landmarks"
    ].transformed_lookuptable

    lookuptable_train_val = {
        "original": lookuptable_train_val_original,
        "transformed": lookuptable_train_val_transformed,
    }

    # get min transform and max transform from train + val lookuptable
    transform_min = lookuptable_train_val_original[transform_min_landmark]["mean"]
    transform_max = lookuptable_train_val_original[transform_max_landmark]["mean"]

    settings = {
        "upper_tangential_quantile": upper_tangential_slope_quantile,
        "lower_tangential_quantile": lower_tangential_slope_quantile,
        "start-landmark": transform_min_landmark,
        "end-landmark": transform_max_landmark,
    }

    if not skip_slopes:
        # Tangential Slopes
        print("Compute tangential slopes of training data set")
        tangential_slopes = compute_tangential_slopes(
            modelEval.inference_model,
            modelEval.train_dataset,
            transform_min,
            transform_max,
        )
        upper_tangential_slope = np.nanquantile(
            tangential_slopes, upper_tangential_slope_quantile
        )
        lower_tangential_slope = np.nanquantile(
            tangential_slopes, lower_tangential_slope_quantile
        )
        tangential_slope_mean = np.nanmean(tangential_slopes)

        # Slice score curve slopes
        print("Compute slice score curve slopes")
        slice_score_curve_slopes = compute_slice_score_curve_slopes(
            modelEval.inference_model,
            modelEval.train_dataset,
            lower_tangential_slope,
            upper_tangential_slope,
            transform_min,
            transform_max,
        )

        slope_mean = np.nanmean(slice_score_curve_slopes)
        slope_std = np.nanstd(slice_score_curve_slopes, ddof=1)
        slope_median = np.nanquantile(slice_score_curve_slopes, 0.5)

        storage = InferenceSettingsStorage(
            slope_mean=slope_mean,
            slope_median=slope_median,
            slope_std=slope_std,
            upper_quantile_tangential_slope=upper_tangential_slope,
            lower_quantile_tangential_slope=lower_tangential_slope,
            tangential_slope_mean=tangential_slope_mean,
            lookuptable_train=lookuptable_train,
            lookuptable_train_val=lookuptable_train_val,
            settings=settings,
        )

    else:
        storage = InferenceSettingsStorage(
            lookuptable_train=lookuptable_train,
            lookuptable_train_val=lookuptable_train_val,
            settings=settings,
        )

    storage.save(model_path)


def compute_tangential_slopes(
    inference_model: InferenceModel,
    train_dataset: BPRDataset,
    transform_min: float,
    transform_max: float,
):
    tangential_slopes = []
    for i in tqdm(range(len(train_dataset))):
        z = train_dataset.z_spacings[i]
        X = train_dataset.get_full_volume(i)
        scores = inference_model.predict_npy_array(X)
        scores = Scores(
            scores,
            z,
            transform_min=transform_min,
            transform_max=transform_max,
            tangential_slope_min=-np.inf,
            tangential_slope_max=np.inf,
        )
        tangential_slopes += list(scores.slopes)

    return np.array(tangential_slopes)


def compute_slice_score_curve_slopes(
    inference_model: InferenceModel,
    train_dataset: BPRDataset,
    lower_tangential_slope: float,
    upper_tangential_slope: float,
    transform_min: float,
    transform_max: float,
):
    curve_slopes = []
    for i in tqdm(range(len(train_dataset))):
        z = train_dataset.z_spacings[i]
        X = train_dataset.get_full_volume(i)
        scores = inference_model.predict_npy_array(X)
        scores = Scores(
            scores,
            z,
            transform_min=transform_min,
            transform_max=transform_max,
            tangential_slope_min=lower_tangential_slope,
            tangential_slope_max=upper_tangential_slope,
        )
        curve_slopes.append(scores.a)

    return np.array(curve_slopes)


@dataclass
class InferenceSettingsStorage:
    slope_mean: float = 0.118
    slope_median: float = 0.118
    slope_std: float = 0.012
    tangential_slope_mean: float = 0.113
    upper_quantile_tangential_slope: float = 0.25
    lower_quantile_tangential_slope: float = -0.037
    lookuptable_train: dict = field(default_factory={})
    lookuptable_train_val: dict = field(default_factory={})
    settings: dict = field(default_factory={})

    def __post_init__(self):
        self.json_dict = self.__dict__.copy()

    def save(self, save_path=""):
        with open(save_path + "inference-settings.json", "w") as f:
            json.dump(self.json_dict, f, indent=4)

    def load(self, path=""):
        with open(path, "r") as f:
            json_dict = json.load(f)

        self.__init__(**json_dict)
        self.__post_init__()



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch
import numpy as np
import sys, os
import matplotlib.pyplot as plt
from scipy.ndimage import gaussian_filter

sys.path.append("../../")
from bpreg.utils.linear_transformations import *



class Scores:
    """Scores and additional meta data inforamtion based on the prediction from the Body Part Regression (bpr) model.

    Args:
        scores (list): predictions from the bpr model.
        zspacing (float): zspacing of analzed volume.
        transform_min (float, optional): score which should get mapped to zero. Defaults to np.nan.
        transform_max (float, optional): score which should get mapped to 100. Defaults to np.nan.
        smoothing_sigma (float, optional): Smoothing sigma in mm, for gaussian smoothing of the scores. Defaults to 10.
        tangential_slope_min (float, optional): minimum valid tangential slope. Defaults to -0.037.
        tangential_slope_max (float, optional): maximum valid tangential slope. Defaults to 0.25.
        slope_mean (float, optional): expected slope of slice score curve. Defaults to np.nan.
        background_scores (list, optional): slice score prediction of empty slice. Defaults to [110.83, 6.14].
        r_slope_threshold (float, optional): threshold for declaring the z-spacing as invalid. Defaults to 0.28.
    """

    def __init__(
        self,
        scores: list,
        zspacing: float,
        transform_min: float,
        transform_max: float,
        smoothing_sigma: float = 10,
        tangential_slope_min: float = -0.037,
        tangential_slope_max: float = 0.25,
        slope_mean: float = np.nan,
        background_scores=[110.83, 6.14],
        r_slope_threshold: float = 0.28,
    ):

        scores = np.array(scores).astype(float)
        self.background_scores = background_scores
        self.length = len(scores)
        self.zspacing = zspacing
        self.smoothing_sigma = smoothing_sigma
        self.transform_min = transform_min
        self.transform_max = transform_max
        self.tangential_slope_min = tangential_slope_min
        self.tangential_slope_max = tangential_slope_max
        self.slope_mean = slope_mean
        self.scale = 100
        self.original_values = scores
        self.original_transformed_values = self.transform_scores(scores.copy())
        self.a_original, self.b_original = self.fit_linear_line(x=np.arange(len(self.original_transformed_values)), 
                                                                y=self.original_transformed_values)


        self.values = self.filter_scores(scores)
        self.values = self.transform_scores(self.values)
        self.values = self.smooth_scores(self.values)
        self.set_boundary_indices(self.values)
        self.values = self.remove_outliers(self.values)
        self.valid_region = np.where(~np.isnan(self.values))[0]

        self.z = np.arange(len(scores)) * zspacing
        self.valid_z = self.z[~np.isnan(self.values)]
        self.valid_values = self.values[~np.isnan(self.values)]
        self.a, self.b = self.fit_linear_line(x=self.valid_z, y=self.valid_values)

        # data sanity chekcs
        self.r_slope_threshold = r_slope_threshold
        self.expected_zspacing = self.calculate_expected_zspacing()
        self.r_slope = self.calculate_relative_error_to_expected_slope()
        self.valid_zspacing = self.is_zspacing_valid()
        self.reverse_zordering = self.is_zordering_reverse()

        # define settings
        self.settings = {"transform_min": self.transform_min,
                "transform_max": self.transform_max,
                "slope_mean": self.slope_mean,
                "tangential_slope_min": self.tangential_slope_min,
                "tangential_slope_max": self.tangential_slope_max,
                "r_slope_threshold": self.r_slope_threshold, 
                "smoothing_sigma": self.smoothing_sigma, 
                "background_scores": self.background_scores}

    def __len__(self):
        return len(self.original_values)

    def smooth_scores(self, x):
        smoothed = x.copy()
        not_nan = np.where(~np.isnan(x))
        not_nan_values = x[not_nan]

        # smooth scores
        smoothed[not_nan] = gaussian_filter(
            not_nan_values, sigma=self.smoothing_sigma / self.zspacing
        )

        return np.array(smoothed)

    def transform_scores(self, x):
        if (not np.isnan(self.transform_min)) & (not np.isnan(self.transform_max)):
            transformed = linear_transform(
                x,
                scale=self.scale,
                min_value=self.transform_min,
                max_value=self.transform_max,
            )
            return transformed

    def filter_scores(self, x):
        """Filter predictions of empty slices."""
        for target_score in self.background_scores:
            x[np.round(x, 2) == target_score] = np.nan
        return x

    def set_boundary_indices(self, x):
        min_score = 0
        max_score = 100

        min_boundary_idx = np.nan
        max_boundary_idx = np.nan

        diff_min_score = np.abs(x - min_score)
        diff_max_score = np.abs(x - max_score)

        if np.nanmin(diff_min_score) < 10:
            min_boundary_idx = np.nanargmin(diff_min_score)

        if np.nanmin(diff_max_score) < 10:
            max_boundary_idx = np.nanargmin(diff_max_score)

        self.min_boundary_idx = min_boundary_idx
        self.max_boundary_idx = max_boundary_idx

        if self.min_boundary_idx > self.max_boundary_idx:
            self.min_boundary_idx = np.nan

    def remove_outliers(self, x):
        if len(x) < 2:
            return x

        # get differences. Estimate last difference by copieng previous value
        diffs = np.array(list(np.diff(x)) + [np.diff(x)[-1]])
        self.slopes = diffs / self.zspacing

        # get outlier slopes
        outlier_indices = np.where(
            (self.slopes < self.tangential_slope_min)
            | (self.slopes > self.tangential_slope_max)
        )[0]

        # if unprocessed slice scores increase  
        if self.a_original > 0: 
            # identify if outliers lie before or after boundary index
            outlier_indices_left_tail = outlier_indices[
                np.where(outlier_indices < self.min_boundary_idx)[0]
            ]
            outlier_indices_right_tail = outlier_indices[
                np.where(outlier_indices > self.max_boundary_idx)[0]
            ]

        # if unprocessed slice scores decrease  
        else: 
            # identify if outliers lie before or after boundary index
            outlier_indices_left_tail = outlier_indices[
                np.where(outlier_indices < self.max_boundary_idx)[0]
            ]
            outlier_indices_right_tail = outlier_indices[
                np.where(outlier_indices > self.min_boundary_idx)[0]
            ]

        # set left and right tail with outlier slopes to nan
        if len(outlier_indices_left_tail) > 0:
            x[: np.max(outlier_indices_left_tail)] = np.nan
        if len(outlier_indices_right_tail) > 0:
            x[np.min(outlier_indices_right_tail) :] = np.nan
        return x

    
    def fit_linear_line(self, x, y):
        if len(x) < 2:
            return np.nan, np.nan
        X = np.full((len(x), 2), 1.0, dtype=float)
        X[:, 1] = x
        b, a = np.linalg.inv(X.T @ X) @ X.T @ y
        return a, b

    def calculate_expected_zspacing(self):
        slope_score2index = self.a * self.zspacing
        expected_zspacing = slope_score2index / self.slope_mean

        return expected_zspacing

    def calculate_relative_error_to_expected_slope(self):
        return np.abs(self.a) / self.slope_mean

    def is_zordering_reverse(self):
        if self.a < 0:
            return 1
        return 0

    def is_zspacing_valid(self):
        if np.isnan(self.r_slope):
            return np.nan

        if np.abs(1 - self.r_slope) > self.r_slope_threshold:
            return 0
        return 1



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import datetime
import random, sys
import torch
import cv2
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torchvision.models as models

cv2.setNumThreads(1)

sys.path.append("../../")
from bpreg.evaluation.landmark_mse import LMSE
from bpreg.network_architecture.loss_functions import *
from bpreg.network_architecture.base_model import BodyPartRegressionBase


class BodyPartRegressionResNet(BodyPartRegressionBase):
    def __init__(
        self,
        lr: float = 1e-4,
        lambda_: float = 0,
        alpha: float = 0,
        pretrained: bool = False,
        delta_z_max: float = np.inf,
        loss_order: str = "h",
        beta_h: float = 0.025,
        alpha_h: float = 0.5,
        weight_decay: float = 0,
    ):

        BodyPartRegressionBase.__init__(
            self,
            lr=lr,
            lambda_=lambda_,
            alpha=alpha,
            pretrained=pretrained,
            delta_z_max=delta_z_max,
            loss_order=loss_order,
            beta_h=beta_h,
            alpha_h=alpha_h,
            weight_decay=weight_decay,
        )

        # load resnet base model
        self.fc7_res = nn.Linear(2048, 1)
        self.model = self.get_resnet()

    def get_resnet(self):
        resnet50 = models.resnet50(pretrained=self.pretrained)
        resnet50.conv1 = torch.nn.Conv2d(
            1, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
        )
        resnet50 = torch.nn.Sequential(*(list(resnet50.children())[:-1]))
        return resnet50

    def forward(self, x: torch.Tensor):
        x = F.relu(self.model(x.float()))
        x = x.view(-1, 2048)
        x = self.fc7_res(x)
        return x



import sys, os, json
import nibabel as nib 
sys.path.append("../")
from bpreg.scripts.bpreg_inference import *
from bpreg.settings import * 
# test if json files are craeted


def test_inference(): 
    """Expensive test to test based on 10 nifti files in the data/test_cases 
    folder if the inference method runs and creates json files. 
    """

    input_path = os.path.join(MAIN_PATH, "data/test_cases/")
    output_path = os.path.join(MAIN_PATH, "data/test_results/")

    nifti_files = len(os.listdir(input_path))

    # remove existing files in the output path 
    for file in os.listdir(output_path): 
        filepath = output_path + file
        if "test_results" in filepath: os.remove(filepath)

    # run 
    bpreg_inference(input_path, output_path, plot=True)

    # Test creation of json files
    json_output_files = [f for f in os.listdir(output_path) if f.endswith(".json")]
    assert len(json_output_files) == nifti_files

    # Test tags 
    with open(output_path + json_output_files[0]) as f: 
        x = json.load(f)

    assert "settings" in x.keys() 
    assert "cleaned slice scores" in x.keys() 
    assert "unprocessed slice scores" in x.keys()
    assert "body part examined" in x.keys()
    assert "body part examined tag" in x.keys()

    # Check if readme file is saved. 
    assert "README.md" in os.listdir(output_path)

def test_plot_scores_in_json_files(): 
    output_path = os.path.join(MAIN_PATH, "data/test_results/")
    plot_scores_in_json_files(output_path)
    
    assert len([f for f in os.listdir(output_path) if f.endswith(".json")]) == len([f for f in os.listdir(output_path) if f.endswith(".png")]) 


def test_npy_inference(): 
    """ Test if npy inference is equal to nifti inference. 
    """
    input_path = os.path.join(MAIN_PATH, "data/test_cases/")
    input_path += os.listdir(input_path)[0]

    model = InferenceModel()
    metadata1 = model.nifti2json(input_path, "")

    img = nib.load(input_path)
    pixel_spacings = img.header.get_zooms()
    X = img.get_fdata(dtype=np.float32)
    metadata2 = model.npy2json(X, "", pixel_spacings=pixel_spacings)

    assert metadata1 == metadata2


if __name__ == "__main__": 
    test_plot_scores_in_json_files()


import numpy as np
from collections import Counter

from bpreg.score_processing.scores import Scores
from bpreg.score_processing.bodypartexamined_dict import get_scoreDict

BODY_PARTS_INCLUDED = {
    "PELVIS": ["pelvis_start", "femur_end", "pelvis_end"],
    "ABDOMEN": [
        "L5",
        "L4",
        "L3",
        "L2",
        "L1",
        "Th12",
        "Th11",
        "Th10",
        "Th9",
    ],
    "CHEST": [
        "Th12",
        "Th11",
        "Th10",
        "Th9",
        "Th8",
        "Th7",
        "Th6",
        "Th5",
        "Th4",
        "Th3",
        "Th2",
        "Th1",
    ],
    "NECK": ["Th3", "Th2", "Th1", "C7", "C6", "C5", "C4", "C3", "C2"],
    "HEAD": ["C3", "C2", "C1", "eyes_end"],
}

MIN_PRESENT_LANDMARKS = {"PELVIS": 2, "ABDOMEN": 7, "CHEST": 9, "NECK": 6, "HEAD": 3}

DISTINCT_BODY_PARTS = {
    "PELVIS": ["pelvis_start", "pelvis_end"],
    "ABDOMEN": ["pelvis_end", "Th9"],
    "CHEST": ["Th9", "Th2"],
    "NECK": ["Th2", "C3"],
    "HEAD": ["C3", "head_end"],
}


class BodyPartExaminedTag:
    def __init__(
        self,
        lookuptable,
        body_parts_included=BODY_PARTS_INCLUDED,
        distinct_body_parts=DISTINCT_BODY_PARTS,
        min_present_landmarks=MIN_PRESENT_LANDMARKS,
        zrange_threshold=100,  # in mm
        ignore_invalid_z: bool=False, 
    ):
        self.ignore_invalid_z = ignore_invalid_z
        self.body_parts_included = body_parts_included
        self.min_present_landmarks = min_present_landmarks
        self.distinct_body_parts = distinct_body_parts
        self.lookuptable = lookuptable
        self.zrange_threshold = zrange_threshold

        self.body_parts_included_scores = get_scoreDict(
            self.body_parts_included, lookuptable
        )
        self.distinct_body_parts_scores = get_scoreDict(
            self.distinct_body_parts, lookuptable
        )

    def estimate_tag(self, scores: Scores):
        # if z-spacing is invalid return NONE as body part examined
        if (scores.valid_zspacing == 0) and not self.ignore_invalid_z:
            return "NONE"

        # check if z-range is greater than 100 mm
        zrange = len(scores.original_values) * scores.zspacing

        tag = self.get_bodypartexamined_from_volume(scores.values)
        if zrange < self.zrange_threshold or (isinstance(tag, float) and np.isnan(tag)):
            tag = self.get_most_frequent_bodypartexamined_in_slices(scores.values)

        return tag

    def is_landmark_present(self, scores: np.array, landmark_score: float) -> bool:
        smaller_scores = np.where(scores < landmark_score)[0]
        bigger_scores = np.where(scores > landmark_score)[0]

        if (len(smaller_scores) > 0) and (len(bigger_scores) > 0):
            return True
        return False

    def is_bodypart_present(
        self, scores: np.array, landmark_scores, landmark_threshold
    ) -> bool:
        landmarks_inside = 0
        for landmark_score in landmark_scores:
            if self.is_landmark_present(scores, landmark_score):
                landmarks_inside += 1

        if landmarks_inside >= landmark_threshold:
            return True
        return False

    def is_score_in_body_part(self, score, bodypart) -> bool:
        bodypart_boundaries = self.distinct_body_parts_scores[bodypart]
        if (score > bodypart_boundaries[0]) & (score < bodypart_boundaries[1]):
            return True
        return False

    def get_most_frequent_bodypartexamined_in_slices(self, scores: np.array) -> str:
        bodyparts_for_slices = []
        for score in scores:
            for bodypart in self.distinct_body_parts_scores:
                if self.is_score_in_body_part(score, bodypart):
                    bodyparts_for_slices.append(bodypart)
                    continue

        if len(bodyparts_for_slices) == 0:
            return "NONE"

        return most_frequent(bodyparts_for_slices)

    def get_bodypartexamined_from_volume(self, scores: np.array) -> str:
        bodyparts_included = []
        for bodypart, landmark_scores in self.body_parts_included_scores.items():
            threshold = self.min_present_landmarks[bodypart]
            if self.is_bodypart_present(scores, landmark_scores, threshold):
                bodyparts_included.append(bodypart)

        return self.join_bodyparts_included(bodyparts_included)

    def join_bodyparts_included(self, bodyparts_list: list):
        if len(bodyparts_list) == 0:
            return np.nan

        if len(bodyparts_list) == 1:
            return bodyparts_list[0]

        if "HEAD" in bodyparts_list:
            if "PELVIS" in bodyparts_list:
                return "HEAD-NECK-CHEST-ABDOMEN-PELVIS"
            elif "ABDOMEN" in bodyparts_list:
                return "HEAD-NECK-CHEST-ABDOMEN"
            elif "CHEST" in bodyparts_list:
                return "HEAD-NECK-CHEST"
            else:
                return "HEAD-NECK"

        if "NECK" in bodyparts_list:
            if "PELVIS" in bodyparts_list:
                return "NECK-CHEST-ABDOMEN-PELVIS"
            elif "ABDOMEN" in bodyparts_list:
                return "NECK-CHEST-ABDOMEN"
            else:
                return "NECK-CHEST"

        if "CHEST" in bodyparts_list:
            if "PELVIS" in bodyparts_list:
                return "CHEST-ABDOMEN-PELVIS"
            else:
                return "CHEST-ABDOMEN"

        return "ABDOMEN-PELVIS"


def most_frequent(List):
    occurence_count = Counter(List)
    return occurence_count.most_common(1)[0][0]



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import os, sys
import torch
import json, pickle
import argparse

sys.path.append("../../")

from bpreg.preprocessing.nifti2npy import Nifti2Npy
from bpreg.network_architecture.bpr_model import BodyPartRegression
from bpreg.score_processing import Scores, BodyPartExaminedDict
from bpreg.settings.settings import * 
from bpreg.settings.model_settings import ModelSettings
from bpreg.score_processing.bodypartexamined_tag import *
from bpreg.utils.json_parser import * 
from bpreg.scripts.initialize_pretrained_model import initialize_pretrained_model


from dataclasses import dataclass
from tqdm import tqdm



class InferenceModel:
    """
    Body Part Regression Model for inference purposes.

    Args:
        base_dir (str]): Path which includes model related file.
        Structure of base_dir:
        base_dir/
            model.pt - includes model
            settings.json - includes mean slope and mean slope std
            lookuptable.json - includes lookuptable as reference
        device (str, optional): [description]. "cuda" or "cpu"
    """

    def __init__(self, base_dir: str=DEFAULT_MODEL, gpu: bool = 1):

        self.base_dir = base_dir
        self.device = "cpu"
        if gpu:
            self.device = "cuda"

        self.model = load_model(base_dir, device=self.device)
        self.load_inference_settings()

        self.n2n = Nifti2Npy(
            target_pixel_spacing=3.5, min_hu=-1000, max_hu=1500, size=128
        )

    def load_inference_settings(self):

        path = self.base_dir + "inference-settings.json"
        if not os.path.exists(path):
            print("WARNING: For this model, no inference settings can be load!")
            return

        with open(path, "rb") as f:
            settings = json.load(f)

        # use for inference the lookuptable from all predictions
        # of the annotated landmarks in the train- and validation-dataset
        self.lookuptable_original = settings["lookuptable_train_val"]["original"]
        self.lookuptable = settings["lookuptable_train_val"]["transformed"]

        self.start_landmark = settings["settings"]["start-landmark"]
        self.end_landmark = settings["settings"]["end-landmark"]

        self.transform_min = self.lookuptable_original[self.start_landmark]["mean"]
        self.transform_max = self.lookuptable_original[self.end_landmark]["mean"]

        self.slope_mean = settings["slope_mean"]
        self.tangential_slope_min = settings["lower_quantile_tangential_slope"]
        self.tangential_slope_max = settings["upper_quantile_tangential_slope"]

    def predict_tensor(self, tensor, n_splits=200):
        scores = []
        n = tensor.shape[0]
        slice_splits = list(np.arange(0, n, n_splits))
        slice_splits.append(n)

        with torch.no_grad():
            self.model.eval()
            self.model.to(self.device)
            for i in range(len(slice_splits) - 1):
                min_index = slice_splits[i]
                max_index = slice_splits[i + 1]
                score = self.model(tensor[min_index:max_index, :, :, :].to(self.device))
                scores += [s.item() for s in score]

        scores = np.array(scores)
        return scores

    def predict_npy_array(self, x, n_splits=200):
        x_tensor = torch.tensor(x[:, np.newaxis, :, :]).to(self.device)
        scores = self.predict_tensor(x_tensor, n_splits=n_splits)
        return scores

    def predict_nifti(self, nifti_path: str):
        # get nifti file as tensor
        try: 
            x, pixel_spacings = self.n2n.preprocess_nifti(nifti_path)
        except: x, pixel_spacings = np.nan, np.nan 

        if isinstance(x, float) and np.isnan(x):
            print(
                f"WARNING: File {nifti_path.split('/')[-1]} can not be converted to a 3-dimensional volume ",
                f"of the size {self.n2n.size}x{self.n2n.size}xz",
            )
            return np.nan

        x = np.transpose(x, (2, 0, 1))[:, np.newaxis, :, :]
        x_tensor = torch.tensor(x)
        x_tensor.to(self.device)

        # predict slice-scores
        scores = self.predict_tensor(x_tensor)
        return self.parse_scores(scores, pixel_spacings[2])

    def parse_scores(self, scores_array, pixel_spacing):

        scores = Scores(
            scores_array,
            pixel_spacing,
            transform_min=self.lookuptable_original[self.start_landmark]["mean"],
            transform_max=self.lookuptable_original[self.end_landmark]["mean"],
            slope_mean=self.slope_mean,
            tangential_slope_min=self.tangential_slope_min,
            tangential_slope_max=self.tangential_slope_max,
        )
        return scores
        

    def npy2json(self, X_: np.array, output_path: str, pixel_spacings: tuple, axis_ordering=(0, 1, 2), ignore_invalid_z: bool =False):
        """
        Method to predict slice scores from numpy arrays (in Hounsfiel dunits). 
        Converts plain numpy array to numpy arrays which can be used by the DEFAULT_MODEL to predict the slice scores.n

        Args:
            X (np.array): matrix of CT volume in Hounsfield units. 
            output_path (str): output path to save json file
            pixel_spacing (tuple): pixel spacing in x, y and z direction. 
            axis_ordering (tuple): Axis ordering of CT volume. (0,1,2) is equivalent to the axis ordering xyz.
            ignore_invalid_z (bool): If true, than invalid z-spacing will be ignored for predicting the body part examined and not NONE will be given back. 
        """
        X = self.n2n.preprocess_npy(X_, pixel_spacings, axis_ordering=axis_ordering) 

        # convert axis ordering to zxy
        X = X.transpose(2, 0, 1)
        
        slice_scores = self.predict_npy_array(X)
        slice_scores = self.parse_scores(slice_scores, pixel_spacings[2])
        data_storage = VolumeStorage(slice_scores, self.lookuptable, ignore_invalid_z=ignore_invalid_z)
        if len(output_path) > 0:
            data_storage.save_json(output_path)
        return data_storage.json

    def nifti2json(self, nifti_path: str, output_path: str="", stringify_json: bool=False, ignore_invalid_z: bool =False):
        """
        Main method to convert NIFTI CT volumes int JSON meta data files. 
        Args:
            nifti_path (str): path of input NIFTI file
            output_path (str): output path to save JSON file
            stringify_json (bool): Set it to true for Kaapana JSON format 
            axis_ordering (tuple): Axis ordering of CT volume. (0,1,2) is equivalent to the axis ordering xyz.
            ignore_invalid_z (bool): If true, than invalid z-spacing will be ignored for predicting the body part examined and not NONE will be given back. 
        """
        slice_scores = self.predict_nifti(nifti_path)
        if isinstance(slice_scores, float) and np.isnan(slice_scores):
            return np.nan

        data_storage = VolumeStorage(slice_scores, self.lookuptable, ignore_invalid_z=ignore_invalid_z)
        if len(output_path) > 0:
            data_storage.save_json(output_path, stringify_json=stringify_json)
        return data_storage.json


@dataclass
class VolumeStorage:
    """Body part metadata for one volume

    Args:
        scores (Scores): predicted slice scores
        lookuptable (dict): reference table which contains expected scores for anatomies
        body_parts ([type], optional): dictionary to define the body parts for the tag: "body part examined". Defaults to BODY_PARTS.
        body_parts_included ([type], optional): dictionary to calculate the "body part examined tag". Defaults to BODY_PARTS_INCLUDED.
        distinct_body_parts ([type], optional): dictionary to calculate the "body part examined tag". Defaults to DISTINCT_BODY_PARTS.
        min_present_landmarks ([type], optional): dictionary to calculate the "body part examined rtag". Defaults to MIN_PRESENT_LANDMARKS.
    """
    def __init__(self, 
                 scores: Scores, 
                 lookuptable: dict, 
                 body_parts=BODY_PARTS, 
                 body_parts_included=BODY_PARTS_INCLUDED, 
                 distinct_body_parts=DISTINCT_BODY_PARTS, 
                 min_present_landmarks=MIN_PRESENT_LANDMARKS, 
                 ignore_invalid_z: bool=False, 
                 ):

        self.ignore_invalid_z = ignore_invalid_z
        self.body_parts=body_parts
        self.body_parts_included = body_parts_included
        self.distinct_body_parts = distinct_body_parts
        self.min_present_landmarks = min_present_landmarks

        self.cleaned_slice_scores = list(scores.values.astype(np.float64))
        self.z = list(scores.z.astype(np.float64))
        self.unprocessed_slice_scores = list(
            scores.original_transformed_values.astype(np.float64)
        )
        self.lookuptable = lookuptable

        self.zspacing = float(scores.zspacing)  # .astype(np.float64)
        self.reverse_zordering = float(scores.reverse_zordering)
        self.valid_zspacing = float(scores.valid_zspacing)
        self.expected_slope = float(scores.slope_mean)
        self.observed_slope = float(scores.a)
        self.expected_zspacing = float(scores.expected_zspacing)
        self.r_slope = float(scores.r_slope)
        self.bpe = BodyPartExaminedDict(lookuptable, body_parts=self.body_parts)
        self.bpet = BodyPartExaminedTag(lookuptable,         
                                        body_parts_included=self.body_parts_included,
                                        distinct_body_parts=self.distinct_body_parts,
                                        min_present_landmarks=self.min_present_landmarks, 
                                        ignore_invalid_z=self.ignore_invalid_z)

        self.settings = {"slice score processing": scores.settings, 
                            "body part examined dict": self.body_parts, 
                           "body part examined tag": {
                                "body parts included": self.body_parts_included, 
                                "distinct body parts": self.distinct_body_parts, 
                                "min present landmarks": self.min_present_landmarks}}

        self.json = {
            "cleaned slice scores": self.cleaned_slice_scores,
            "z": self.z,
            "unprocessed slice scores": self.unprocessed_slice_scores,
            "body part examined": self.bpe.get_examined_body_part(
                self.cleaned_slice_scores
            ),
            "body part examined tag": self.bpet.estimate_tag(scores),
            "look-up table": self.lookuptable,
            "reverse z-ordering": self.reverse_zordering,
            "valid z-spacing": self.valid_zspacing,
            "expected slope": self.expected_slope,
            "observed slope": self.observed_slope,
            "slope ratio": self.r_slope,
            "expected z-spacing": self.expected_zspacing,
            "z-spacing": self.zspacing,
            "settings": self.settings
        }

    def save_json(self, output_path: str, stringify_json=False):
        """Store data in json file 

        Args:
            output_path (str): save path for json file 
            stringify_json (bool, optional): if True, stringify output of parameters and
            convert json file to a Kaapana friendly format
        """
        data = self.json 
        if stringify_json: 
            data = parse_json4kaapana(data)

        with open(output_path, "w") as f:
            json.dump(data, f, indent=4)


def load_model(
    base_dir, model_file="model.pt", config_file="config.json", device="cuda"
):
    # load public model, if it does not exist locally
    if (base_dir == DEFAULT_MODEL) & ~os.path.exists(base_dir):
        initialize_pretrained_model()

    config_filepath = base_dir + config_file
    model_filepath = base_dir + model_file

    config = ModelSettings()
    config.load(path=config_filepath)

    model = BodyPartRegression(alpha=config.alpha, lr=config.lr)
    model.load_state_dict(torch.load(model_filepath, map_location=torch.device(device)), strict=False)
    model.eval()
    model.to(device)

    return model


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--i", default="")
    parser.add_argument("--o", default="")
    parser.add_argument("--g", default=1)

    value = parser.parse_args()
    ipath = value.i
    opath = value.o
    gpu = value.g

    base_dir = "../../src/models/private_bpr_model/"
    model = InferenceModel(base_dir, gpu=gpu)

    data_path = "../../data/test_cases/"
    nifti_paths = [
        data_path + f for f in os.listdir(data_path) if f.endswith(".nii.gz")
    ]
    for nifti_path in tqdm(nifti_paths):
        output_path = nifti_path.replace("test_cases", "test_results").replace(
            ".nii.gz", ".json"
        )
        model.nifti2json(nifti_path, output_path)



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import requests, zipfile, os
from bpreg.settings.settings import MAIN_PATH


PUBLIC_MODEL_URL = (
    "https://zenodo.org/record/5113483/files/public_bpr_model.zip?download=1"
)

SAVE_PUBLIC_MODEL_DIR_PATH =  os.path.join(
    MAIN_PATH, "src/models/"
)
SAVE_PUBLIC_MODEL_PATH = os.path.join(SAVE_PUBLIC_MODEL_DIR_PATH, "public_inference_model.zip")



def initialize_pretrained_model():
    # Download public model from zenodo, for inference
    print(
        "Download publicly available body part regression model from Zenodo:\n",
        "https://zenodo.org/record/5113483",
    )

    with requests.get(PUBLIC_MODEL_URL, stream="True") as r:
        r.raise_for_status()
        if not os.path.exists(os.path.join(MAIN_PATH, "src")): os.mkdir(os.path.join(MAIN_PATH, "src"))
        if not os.path.exists(os.path.join(MAIN_PATH, "src/models/")): os.mkdir(os.path.join(MAIN_PATH, "src/models/"))

        with open(SAVE_PUBLIC_MODEL_PATH, "wb") as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)

    # unzip public model
    save_path = os.path.dirname(SAVE_PUBLIC_MODEL_PATH)
    with zipfile.ZipFile(SAVE_PUBLIC_MODEL_PATH, "r") as f:
        f.extractall(save_path)

    # remove file
    os.remove(SAVE_PUBLIC_MODEL_PATH)


if __name__ == "__main__":
    initialize_pretrained_model()



from bpreg.score_processing.scores import Scores
from bpreg.score_processing.landmark_scores import LandmarkScores, LandmarkScoreBundle
from bpreg.score_processing.bodypartexamined_dict import BodyPartExaminedDict



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random, pickle, datetime, os, sys, cv2, json
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import argparse
import albumentations as A
from scipy.stats import pearsonr
from tqdm import tqdm


import torch, torchvision
import torch.nn as nn
import torchvision.models as models
import pytorch_lightning as pl
from pytorch_lightning import seed_everything
from pytorch_lightning.loggers import TensorBoardLogger


sys.path.append("../")
from bpreg.network_architecture.bpr_model import BodyPartRegression
from bpreg.network_architecture.ssbr_model import SSBR
from bpreg.utils.training_utils import *
from bpreg.settings import ModelSettings

cv2.setNumThreads(1)
np.seterr(divide="ignore", invalid="ignore")

def train_config(config: ModelSettings): 
    seed_everything(config.random_seed)

    # load data
    df = get_dataframe(config)

    if config.model == "SSBR":
        train_dataloader, val_dataloader, _ = data_preprocessing_ssbr(df, config)
        model = SSBR(alpha=config.alpha, lr=config.lr)

    else:
        train_dataset, val_dataset, _ = get_datasets(config, df)
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=config.batch_size,
            num_workers=22,
            shuffle=config.shuffle_train_dataloader,
        )

        val_dataloader = torch.utils.data.DataLoader(
            val_dataset, batch_size=config.batch_size, num_workers=22
        )

        # run model
        run_fast_dev(config, train_dataloader, val_dataloader)

        model = BodyPartRegression(
            alpha=config.alpha,
            lr=config.lr,
            lambda_=config.lambda_,
            alpha_h=config.alpha_h,
            beta_h=config.beta_h,
            base_model=config.base_model,
            loss_order=config.loss_order,
        )

    logger_uar = TensorBoardLogger(save_dir=config.save_dir, name=config.model_name)
    trainer = pl.Trainer(
        gpus=1,
        max_epochs=config.epochs,
        precision=16,
        logger=logger_uar,
        deterministic=config.deterministic,
        accumulate_grad_batches=int(config.effective_batch_size / config.batch_size),
    )

    trainer.fit(model, train_dataloader, val_dataloader)
    if config.save_model: save_model(model, config, path=logger_uar.log_dir + "/")

def train_json(json_path: str):
    config = ModelSettings()
    config.load(json_path)
    print(config)
    train_config(config)



def train_json_list(config_filepaths: list):
    """Train for each config in config_filepaths a model.

    Args:
        config_filepaths (list): list of paths to config-files
    """

    # run code for different configurations
    for filepath in config_filepaths:
        train_json(filepath)


if __name__ == "__main__":

    # get arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("--list", nargs="+", default=[])
    parser.add_argument(
        "--path",
        default="/home/AD/s429r/Documents/Code/bodypartregression/src/configs/local/test/",
    )

    value = parser.parse_args()
    config_filenames = value.list
    config_filepath = value.path

    # if no filenames are defined use all files from path
    if len(config_filenames) == 0:
        config_filenames = [
            f for f in np.sort(os.listdir(config_filepath)) if f.endswith(".json")
        ]

    config_filepaths = [config_filepath + file for file in config_filenames]
    train_json_list(config_filepaths)



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pickle, sys, cv2
import pandas as pd
import numpy as np
import torch, os
import pytorch_lightning as pl

sys.path.append("../")
from bpreg.network_architecture.bpr_model import BodyPartRegression
from bpreg.dataset.bpr_dataset import BPRDataset
from bpreg.dataset.ssbr_dataset import SSBRDataset
from bpreg.score_processing import LandmarkScores
from bpreg.settings import ModelSettings

# np.seterr(divide="ignore", invalid="ignore")
cv2.setNumThreads(1)


def run_fast_dev(
    config: ModelSettings,
    train_dataloader: torch.utils.data.DataLoader,
    val_dataloader: torch.utils.data.DataLoader,
):
    """Test model training

    Args:
        config (dict): configs for training
        train_dataloader (torch.utils.data.DataLoader): train dataloader
        val_dataloader (torch.utils.data.DataLoader): val dataloader
    """
    model = BodyPartRegression(
        alpha=config.alpha,
        lr=config.lr,
        lambda_=config.lambda_,
        alpha_h=config.alpha_h,
        beta_h=config.beta_h,
        base_model=config.base_model,
        loss_order=config.loss_order,
    )
    trainer_dev = pl.Trainer(gpus=1, fast_dev_run=True, precision=16)
    trainer_dev.fit(model, train_dataloader, val_dataloader)


def save_model(model, config: ModelSettings, path):
    print("save model at: ", path)
    config.save(save_path=os.path.join(path, "config.json"))

    if config.save_model:
        torch.save(model.state_dict(), path + "model.pt")

    # save lookuptable
    df_landmarks = pd.read_excel(config.landmark_path, sheet_name="landmarks-train", engine="openpyxl")
    lscores = LandmarkScores(
        config.data_path,
        df_landmarks,
        model,
        landmark_start=np.nan,
        landmark_end=np.nan,
    )
    lscores.save_lookuptable(filepath=os.path.join(path , "lookuptable.json"))


def data_preprocessing_ssbr(df: pd.DataFrame, config: ModelSettings):
    """Obtain datasets for ssbr model

    Args:
        df (pd.DataFrame): dataframe with information to train/val/test split and z-spcaings
        config (dict): config for training

    Returns:
        train-, val- & test-dataloader
    """
    train_filenames = df.loc[df.train_data == 1, "filename"].values
    val_filenames = df.loc[df.val_data == 1, "filename"].values
    test_filenames = df.loc[df.test_data == 1, "filename"].values

    train_zspacings = df.loc[df.train_data == 1, "pixel_spacingz"].values
    val_zspacings = df.loc[df.val_data == 1, "pixel_spacingz"].values
    test_zspacings = df.loc[df.test_data == 1, "pixel_spacingz"].values

    train_dataset = SSBRDataset(
        data_path=config.data_path,
        filenames=train_filenames,
        z_spacings=train_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-train",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    val_dataset = SSBRDataset(
        data_path=config.data_path,
        filenames=val_filenames,
        z_spacings=val_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-val",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    test_dataset = SSBRDataset(
        data_path=config.data_path,
        filenames=test_filenames,
        z_spacings=test_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-test",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    train_dataloader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=config.batch_size,
        num_workers=20,
        shuffle=config.shuffle_train_dataloader,
    )

    val_dataloader = torch.utils.data.DataLoader(
        val_dataset, batch_size=config.batch_size, num_workers=20
    )

    test_dataloader = torch.utils.data.DataLoader(
        test_dataset, batch_size=config.batch_size, num_workers=20
    )

    return train_dataloader, val_dataloader, test_dataloader


def get_dataframe(config: ModelSettings):
    """Get dataframe, which includes all information about train/val/test split
    and z-spacings

    Args:
        config (dict): dictionary contains all information to train a model

    Returns:
        df: dataframe with train/val/test information + z-spacing information
    """
    df = pd.read_excel(config.df_data_source_path, engine="openpyxl")

    # only use volumes with more than 30 slices
    if "z" in df.columns: ddf = df[(df["z"] >= 30)]
    return df


def get_datasets(config: ModelSettings, df: pd.DataFrame):
    """Obtain train, val and test datasets for training

    Args:
        config (dict): configs for training
        df (pd.DataFrame): dataframe contains information to train/val/test split + z-spcaings

    Returns:
        train, val and test dataset
    """
    train_filenames = df.loc[df.train_data == 1, "filename"].values
    val_filenames = df.loc[df.val_data == 1, "filename"].values
    test_filenames = df.loc[df.test_data == 1, "filename"].values

    train_zspacings = df.loc[df.train_data == 1, "pixel_spacingz"].values
    val_zspacings = df.loc[df.val_data == 1, "pixel_spacingz"].values
    test_zspacings = df.loc[df.test_data == 1, "pixel_spacingz"].values

    train_dataset = BPRDataset(
        data_path=config.data_path,
        filenames=train_filenames,
        z_spacings=train_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-train",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    val_dataset = BPRDataset(
        data_path=config.data_path,
        filenames=val_filenames,
        z_spacings=val_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-val",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    test_dataset = BPRDataset(
        data_path=config.data_path,
        filenames=test_filenames,
        z_spacings=test_zspacings,
        landmark_path=config.landmark_path,
        landmark_sheet_name="landmarks-test",
        random_seed=config.random_seed,
        custom_transform=config.custom_transform,
        albumentation_transform=config.albumentation_transform,
        equidistance_range=config.equidistance_range,
        num_slices=config.num_slices,
    )

    return train_dataset, val_dataset, test_dataset



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import random
import cv2


class rescale_intensity(object):
    def __init__(
        self, low: float, high: float, scale: float = 2, dtype: type = np.uint8
    ):
        self.low = low
        self.high = high
        self.scale = scale
        self.dtype = dtype

    def __call__(self, x):
        x[x < self.low] = self.low
        x[x > self.high] = self.high
        x = x - self.low
        x = x / (self.high - self.low)
        x = x * self.scale
        x = x.astype(dtype=self.dtype)

        return x


class adjust_contrast(object):
    def __init__(
        self,
        alpha_min: float = 0,
        alpha_max: float = 1,
        beta_min: float = 1.5,
        beta_max: float = 3.5,
        p_alpha: float = 0.75,
        p_beta: float = 0.75,
    ):
        self.alpha_min = alpha_min
        self.alpha_max = alpha_max
        self.alpha_diff = alpha_max - alpha_min

        self.beta_min = beta_min
        self.beta_max = beta_max
        self.beta_diff = beta_max - beta_min

        self.p_alpha = p_alpha
        self.p_beta = p_beta

    def __call__(self, x):
        mean = np.mean(x)
        std = np.std(x)

        if random.random() <= self.p_alpha:
            a_rand = random.random() * self.alpha_diff + self.alpha_min
            min_value = mean - a_rand * std
            x = np.where(x < min_value, min_value, x)

        if random.random() <= self.p_beta:
            b_rand = random.random() * self.beta_diff + self.beta_min
            max_value = mean + b_rand * std
            x = np.where(x > max_value, max_value, x)

        return x


class AddFrame:
    def __init__(
        self,
        r_square: float = 0.6,
        r_circle: float = 0.75,
        dimension: int = 128,
        p: float = 0.2,
        fill_value: float = -1.0,
    ):
        self.p = p
        self.dimension = dimension
        self.r_square = r_square
        self.r_circle = r_circle
        self.fill_value = fill_value
        self.square_frame = self.get_square_frame()
        self.circle_frame = self.get_circle_frame()

    def get_square_frame(self):
        irange = int(self.dimension * self.r_square + 0.5)
        if irange % 2 == 0:
            irange += 1

        X = np.full((self.dimension, self.dimension), np.nan)
        X_inside = np.full((irange, irange), 0)
        X[
            self.dimension // 2 - irange // 2 : self.dimension // 2 + irange // 2 + 1,
            self.dimension // 2 - irange // 2 : self.dimension // 2 + irange // 2 + 1,
        ] = X_inside

        return X  # [:, :, np.newaxis]

    def get_circle_frame(self):
        irange = int(self.dimension * self.r_circle + 0.5)
        if irange % 2 == 0:
            irange += 1
        X = np.full((self.dimension, self.dimension), np.nan)
        X_inside = np.full((irange, irange), 0, dtype=float)
        center = irange // 2
        radius = center
        for idx in range(irange):
            for idy in range(irange):

                if (idx - center) ** 2 + (idy - center) ** 2 > radius ** 2:
                    X_inside[idx, idy] = np.nan

        X[
            self.dimension // 2 - irange // 2 : self.dimension // 2 + irange // 2 + 1,
            self.dimension // 2 - irange // 2 : self.dimension // 2 + irange // 2 + 1,
        ] = X_inside

        return X  # [:, :, np.newaxis]

    def __call__(self, x):
        if random.random() < self.p:
            x = x + self.circle_frame
            x[np.isnan(x)] = self.fill_value
            x = np.nan_to_num(x)
        return x


class GaussNoise(object):
    def __init__(
        self,
        std_min: float = 0,
        std_max: float = 5,
        min_value: float = -1,
        max_value: float = 1,
        p: float = 0.5,
    ):
        self.std_min = std_min
        self.std_max = std_max
        self.min_value = min_value
        self.max_value = max_value
        self.p = p

    def __call__(self, x):
        if random.random() < self.p:
            sigma = random.uniform(self.std_min, self.std_max)
            gauss = np.random.normal(0, sigma, x.shape)
            x = x + gauss
            x[x < self.min_value] = self.min_value
            x[x > self.max_value] = self.max_value
        return x


class ShiftHU(object):
    def __init__(
        self,
        limit: float = 2.55,
        max_value: float = 1,
        min_value: float = -1,
        p: float = 0.5,
    ):
        self.limit = limit
        self.p = p
        self.min_value = min_value
        self.max_value = max_value

    def __call__(self, x):
        if random.random() < self.p:
            shift = random.random() * 2 * self.limit - self.limit
            x = np.where(x > 0, x + shift, x)
            x[x < self.min_value] = self.min_value
            x[x > self.max_value] = self.max_value
        return x


class ScaleHU:
    def __init__(
        self,
        scale_delta: float = 0.2,
        min_value: float = -1,
        max_value: float = 1,
        p: float = 1,
    ):
        self.p = p
        self.scale_delta = scale_delta
        self.max_value = max_value
        self.min_value = min_value

    def __call__(self, x):
        if random.random() < self.p:
            factor = np.random.uniform(1 - self.scale_delta, 1 + self.scale_delta)
            x = x * factor
            x[x > self.max_value] = self.max_value
            x[x < self.min_value] = self.min_value

        return x


class RandomGamma:
    def __init__(
        self, gamma_min: float, gamma_max: float, max_value: float = 1, p: float = 0.5
    ):
        self.gamma_min = gamma_min
        self.gamma_max = gamma_max
        self.max_value = max_value
        self.p = p

    def __call__(self, x):
        if random.random() < self.p:
            gamma = np.random.uniform(self.gamma_min, self.gamma_max)
            x = np.power(x, gamma)
            x[x > self.max_value] = self.max_value
            x = x.astype(np.float32)
        return x



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import matplotlib.pyplot as plt
import numpy as np
import json, random
import pandas as pd

from bpreg.settings import ModelSettings
from bpreg.utils.training_utils import * 

class Visualization:
    def __init__(self):
        pass

    def plot_landmarks(
        self,
        score_matrix: np.array,
        expected_scores: np.array,
        landmark_names: list,
        figsize: tuple = (16, 10),
        fontsize: float = 24,
        text_margin_top: float = 0.02,
        text_margin_right: float = 2,
        alpha: float = 0.7,
        ylim: None = None,
        colors: list = [],
    ):
        plt.figure(figsize=figsize)
        max_value = 0

        for idx in range(score_matrix.shape[1]):
            x = score_matrix[:, idx]
            if len(colors) == score_matrix.shape[1]:
                bins, _, _ = plt.hist(
                    x,
                    density=True,
                    alpha=alpha,
                    label=landmark_names[idx],
                    color=colors[idx],
                )
            else:
                bins, _, _ = plt.hist(
                    x,
                    density=True,
                    alpha=alpha,
                    label=landmark_names[idx],
                )

            mean = expected_scores[idx]
            plt.plot(
                [mean, mean],
                [0, 100],
                color="black",
                linestyle="--",
                linewidth=0.5,
            )

            plt.annotate(
                landmark_names[idx].replace("_", "-"),
                xy=(mean - text_margin_right, np.max(bins) + text_margin_top),
                fontsize=fontsize - 2,
            )

            if np.max(bins) > max_value:
                max_value = np.max(bins)

        if not ylim:
            plt.ylim((0, 1.1 * max_value))
        else:
            plt.ylim(ylim)
        plt.xticks(fontsize=fontsize - 2)
        plt.yticks(fontsize=fontsize - 2)
        plt.xlabel("Slice Scores", fontsize=fontsize)
        plt.ylabel("Relative Frequency", fontsize=fontsize)
        plt.xlim((-10, 110))


def grid_plot(
    X: np.ndarray,
    titles: list = [],
    cols: int = 4,
    rows: int = 4,
    save_path: str = "",
    cmap: str = "gray",
    vmin: int = 0,
    vmax: int = 250,
    figsize: tuple = (16, 10),
):
    _, axs = plt.subplots(rows, cols, figsize=figsize)
    idx = 0
    for row in range(rows):
        for col in range(cols):
            axs[row, col].imshow(X[idx], cmap=cmap, vmin=vmin, vmax=vmax)
            if len(titles) == cols * rows:
                axs[row, col].set_title(titles[idx], fontsize=15)
            axs[row, col].set_yticklabels([])
            axs[row, col].set_xticklabels([])
            idx += 1

    if len(save_path) > 0:
        plt.savefig(save_path)
    plt.show()


def plot_data(config: ModelSettings, kind: str="train", cols=3, rows=3): 
    df = get_dataframe(config)
    df = df[~df.pixel_spacingz.isna()]
    train_dataset, val_dataset, test_dataset = get_datasets(config, df)

    datasets = {"train": train_dataset, "validation": val_dataset, "test": test_dataset}
    dataset = datasets[kind]

    X = []
    titles  = []
    for i in range(0, cols*rows): 
        index = random.randint(0, len(dataset)- 1)
        titles.append(index)
        X.append(dataset[index][0][0, :, :])

    grid_plot(X, titles=titles, cols=cols, rows=rows, vmin=-1, vmax=1, figsize=(15, 15))


def plot_scores(filepath, save_path="", fontsize=18): 

    def load_json(filepath): 
        with open(filepath) as f: 
            x = json.load(f)
        return x

    plt.figure(figsize=(12, 6))
    x = load_json(filepath)

    plt.plot(x["z"], x["cleaned slice scores"], label="Cleaned Slice Scores")
    plt.plot(x["z"], x["unprocessed slice scores"], label="Unprocessed Slice Scores", linestyle="--")
    
    try: 
        min_score = np.nanmin(x["unprocessed slice scores"])
        max_score = np.nanmax(x["unprocessed slice scores"])
        dflandmarks = pd.DataFrame(x["look-up table"]).T
        landmarks = dflandmarks[(dflandmarks["mean"] > min_score) & (dflandmarks["mean"] < max_score)].sort_values(by="mean", ascending=False)
        for landmark, row in landmarks.iloc[[0, -1]].iterrows(): 
            plt.plot([0, np.nanmax(x["z"])], [row["mean"], row["mean"]], 
                    linestyle = ":", color="black", linewidth=0.8)
            plt.text(5, row["mean"]+1, landmark, fontsize=fontsize - 4,    
                    bbox=dict(boxstyle="square",
                    fc=(1., 1, 1),
                    ))
    except: pass

    plt.xlabel("Height [mm]", fontsize=fontsize)
    plt.ylabel("Slice Scores", fontsize=fontsize)
    plt.xticks(fontsize=fontsize-2)
    plt.yticks(fontsize=fontsize-2)
    plt.legend(loc=1, fontsize=fontsize)
    if np.nanmax(x["z"]) != 0: plt.xlim((0, np.nanmax(x["z"])))
    
    filename = filepath.split("/")[-1]
    if len(filename) > 60: 
        plt.title(f"{filename[:60]}...\n", fontsize=fontsize-2)
    else: 
        plt.title(filename + "\n", fontsize=fontsize-2)
    if len(save_path) > 0: 
        plt.savefig(save_path)
    else: 
        plt.show()



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import random, sys
import torch
import cv2
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models

cv2.setNumThreads(1)

sys.path.append("../../")
from bpreg.network_architecture.loss_functions import *
from bpreg.network_architecture.base_model import BodyPartRegressionBase


class SSBR(BodyPartRegressionBase):
    """[summary]
    SSBR model from Yan at. al with corrected classification order loss (no nan-values can be calculated).
    Args:
        BodyPartRegressionBase ([type]): [description]
    """

    def __init__(self, lr=1e-4, alpha=0):

        super().__init__()
        BodyPartRegressionBase.__init__(
            self,
            lr=lr,
            lambda_=0,
            alpha=alpha,
            pretrained=False,
            delta_z_max=np.inf,
            loss_order="c",
            beta_h=np.nan,
            alpha_h=np.nan,
            weight_decay=0,
        )

        # load vgg base model
        self.conv6 = nn.Conv2d(
            512, 512, 1, stride=1, padding=0
        )  # in_channel, out_channel, kernel_size
        self.fc7 = nn.Linear(512, 1)
        self.model = self.get_vgg()

    def get_vgg(self):
        vgg16 = models.vgg16(pretrained=self.pretrained)
        vgg16.features[0] = torch.nn.Conv2d(
            1, 64, kernel_size=3, stride=1, padding=1, bias=False
        )

        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        vgg16.to(device)

        return vgg16.features

    def forward(self, x: torch.Tensor):
        x = self.model(x.float())
        x = F.relu(self.conv6(x))
        x = torch.mean(x, axis=(2, 3))
        x = x.view(-1, 512)
        x = self.fc7(x)
        return x

    def loss(self, scores_pred, slice_indices, _):
        loss_order = self.loss_order(scores_pred, _)
        ldist = self.alpha * self.loss_dist(scores_pred)
        loss = loss_order + ldist
        return loss, loss_order, ldist, 0

    def loss_dist(self, scores_pred):
        scores_diff = scores_pred[:, 1:] - scores_pred[:, :-1]
        loss = self.l1loss(scores_diff[:, 1:], scores_diff[:, :-1])
        return loss



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys, os
import numpy as np
import pandas as pd
import torch
import pytorch_lightning as pl

sys.path.append("../../")
from bpreg.dataset.base_dataset import get_slices
from bpreg.evaluation.landmark_mse import LMSE
from bpreg.evaluation.accuracy import Accuracy
from bpreg.utils.linear_transformations import *
from bpreg.settings.settings import *

import json


class LandmarkScores:
    def __init__(
        self,
        data_path: str,
        df: pd.DataFrame,
        model: pl.LightningModule,
        device: str = "cuda",
        drop_cols=["val", "train", "test"],
        landmark_start: str = "pelvis_start",
        landmark_end: str = "eyes_end",
    ):

        self.data_path = data_path
        self.device = device
        self.model = model

        # expect filenames to be with or without .npy ending
        self.filenames = [
            f.replace(".npy", "") + ".npy" for f in df["filename"] if isinstance(f, str)
        ]
        self.filepaths = [os.path.join(data_path, f) for f in self.filenames]
        self.landmark_names = [
            l for l in df.columns if not l in drop_cols + ["filename"]
        ]

        self.index_matrix = np.array(
            df.drop(["filename"] + drop_cols, axis=1, errors="ignore")
        )
        self.score_matrix = self.create_score_matrix()

        self.expected_scores = np.nanmean(self.score_matrix, axis=0)
        self.expected_scores_std = np.nanstd(self.score_matrix, axis=0, ddof=1)

        self.expected_scores_transformed = self.transform(self.expected_scores.copy())
        self.score_matrix_transformed = self.transform(self.score_matrix.copy())

        self.lookuptable = self.create_lookuptable()
        if isinstance(landmark_start, float) and np.isnan(landmark_start):
            landmark_start = get_min_keyof_lookuptable(self.lookuptable)
        if isinstance(landmark_end, float) and np.isnan(landmark_end):
            landmark_end = get_max_keyof_lookuptable(self.lookuptable)
        self.transformed_lookuptable = transform_lookuptable(
            self.lookuptable, landmark_start=landmark_start, landmark_end=landmark_end
        )

    def create_score_matrix(self):
        with torch.no_grad():
            self.model.eval()
            self.model.to(self.device)
            slice_score_matrix = np.full(self.index_matrix.shape, np.nan)
            for i in np.arange(len(self.index_matrix)):
                filepath = self.filepaths[i]
                idxs = self.index_matrix[i, :]
                not_isnan = np.where(~np.isnan(idxs))
                idxs = idxs[not_isnan].astype(int)
                X = get_slices(filepath, idxs)
                y = self.model.predict_npy_array(X)
                slice_score_matrix[i, not_isnan] = y

        return slice_score_matrix

    def transform(self, x):
        min_value = self.expected_scores[0]
        max_value = self.expected_scores[-1]
        return linear_transform(x, scale=100, min_value=min_value, max_value=max_value)

    def create_lookuptable(self):
        lookuptable = {l: {} for l in self.landmark_names}

        for i, l in enumerate(self.landmark_names):
            lookuptable[l]["mean"] = self.expected_scores[i]
            lookuptable[l]["std"] = self.expected_scores_std[i]
        return lookuptable

    def print_lookuptable(self):
        for landmark, values in self.transformed_lookuptable.items():
            mean = np.round(values["mean"], 3)
            std = np.round(values["std"], 3)
            print(f"{landmark:<15}:\t {mean} +- {std}")

    def save_lookuptable(self, filepath):
        jsonDict = {
            "original": self.lookuptable,
            "transformed": self.transformed_lookuptable,
        }
        with open(filepath, "w") as f:
            json.dump(jsonDict, f, indent=4)


class LandmarkScoreBundle:
    def __init__(
        self,
        data_path,
        landmark_path,
        model,
        landmark_start="pelvis_start",
        landmark_end="eyes_end",
    ):
        df_database = pd.read_excel(landmark_path, sheet_name="database")
        df_train = pd.read_excel(landmark_path, sheet_name="landmarks-train")
        df_val = pd.read_excel(landmark_path, sheet_name="landmarks-val")
        df_test = pd.read_excel(landmark_path, sheet_name="landmarks-test")

        self.dict = {
            "validation": LandmarkScores(
                data_path,
                df_val,
                model,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            "train": LandmarkScores(
                data_path,
                df_train,
                model,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            "test": LandmarkScores(
                data_path,
                df_test,
                model,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            "train+val-all-landmarks": LandmarkScores(
                data_path,
                df_database[(df_database.train == 1) | (df_database.val == 1)],
                model,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            "test-all-landmarks": LandmarkScores(
                data_path,
                df_database[(df_database.test == 1)],
                model,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
        }
        self.lmse = LMSE()
        self.model = model

    def nMSE(self, target="validation", reference="train"):
        lmse, lmse_std = self.lmse.from_matrices(
            self.dict[target].score_matrix, self.dict[reference].score_matrix
        )
        return lmse, lmse_std

    def nMSE_per_volume(self, target="validation", reference="train"):
        lmse, lmse_std = self.lmse.lmse_per_volume_from_matrices(
            self.dict[target].score_matrix, self.dict[reference].score_matrix
        )
        return lmse, lmse_std

    def accuracy(
        self, target_dataset, reference="train", class2landmark=CLASS_TO_LANDMARK_5
    ):
        acc = Accuracy(self.dict[reference].expected_scores, class2landmark)
        myAccuracy = acc.from_dataset(self.model, target_dataset)
        return myAccuracy

    def nMSE_per_landmark(self, target="validation", reference="train"):
        score_matrix = self.dict[target].score_matrix
        reference_matrix = self.dict[reference].score_matrix
        expected_scores = self.dict[reference].expected_scores
        d = self.lmse.get_normalizing_constant(expected_scores)
        landmark_names = self.dict[reference].landmark_names

        lmse_per_lanmdark = {landmark_name: {} for landmark_name in landmark_names}
        lmses, lmses_errors = self.lmse.lmse_per_landmark_from_matrices(
            score_matrix, reference_matrix
        )

        for landmark, lmse, lmse_std in zip(landmark_names, lmses, lmses_errors):
            lmse_per_lanmdark[landmark]["mean"] = lmse
            lmse_per_lanmdark[landmark]["std"] = lmse_std

        return lmse_per_lanmdark


def get_max_keyof_lookuptable(myDict):
    max_key = ""
    max_key_value = -np.inf
    for key in myDict:
        if myDict[key]["mean"] > max_key_value:
            max_key = key
            max_key_value = myDict[key]["mean"]
    return max_key


def get_min_keyof_lookuptable(myDict):
    min_key = ""
    min_key_value = +np.inf
    for key in myDict:
        if myDict[key]["mean"] < min_key_value:
            min_key = key
            min_key_value = myDict[key]["mean"]
    return min_key



import numpy as np
from bpreg.evaluation.accuracy import Accuracy
from bpreg.evaluation.landmark_mse import LMSE


def test_landmark_mse():
    # test normalized mse with full matrices
    tm = np.array([[1, 2, 4], [1.5, 2.2, 3.5], [1, 1.9, 4.1]])
    vm = np.array([[1.1, 2.1, 3.9], [0.8, 2.0, 4.2]])

    normMSE = LMSE()
    mse, mse_std = normMSE.from_matrices(vm, tm)
    
    assert np.round(mse, 1) == 58.7
    assert np.round(mse_std, 1) == 54.1

    # test normalized mse with nan-values
    tm = np.array([[1, 2, 4], [np.nan, 2.2, 3.5], [1, 1.9, 4.1]])
    vm = np.array([[1.1, 2.1, np.nan], [0.8, 2.0, 4.2]])

    mse, mse_std = normMSE.from_matrices(vm, tm)

    assert np.round(mse, 0) == 35
    assert np.round(mse_std, 0) == 26


def test_accuracy_class_initalization():
    estimated_landmark_slice_scores = np.array(
        [-10, -7.5, -6, -5.5, -4, -3.5, -2, 0, 2, 4.5, 6, 8]
    )
    class_to_landmark = {0: [0, 2], 1: [2, 5], 2: [5, 8], 3: [8, 10], 4: [10, 11]}

    acc = Accuracy(estimated_landmark_slice_scores, class_to_landmark)

    # test class parameter
    assert acc.class_to_score_mapping == {
        0: [-10.0, -6.0],
        1: [-6.0, -3.5],
        2: [-3.5, 2.0],
        3: [2.0, 6.0],
        4: [6.0, 8.0],
    }

    estimated_landmark_slice_scores = np.array([-5, -2, 0, 1, 4])
    class_to_landmark = {0: [0, 1], 1: [1, 3], 2: [3, 4]}

    acc = Accuracy(estimated_landmark_slice_scores, class_to_landmark)
    assert acc.class_to_score_mapping == {0: [-5, -2], 1: [-2, 1], 2: [1, 4]}

    # test class prediction


def test_class_prediction():
    estimated_landmark_slice_scores = np.array(
        [-10, -7.5, -6, -5.5, -4, -3.5, -2, 0, 2, 4.5, 6, 8]
    )
    class_to_landmark = {0: [0, 2], 1: [2, 5], 2: [5, 8], 3: [8, 10], 4: [10, 11]}

    slice_scores = np.array(
        [-4.0, -3.5, -3.0, -2.5, -2.0, -1.5, -1.0, -0.5, 0.0, 0.5, 1.0, 1.5, 2.0, 2.5]
    )
    predicted_classes = np.array(
        [1.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0]
    )
    acc = Accuracy(estimated_landmark_slice_scores, class_to_landmark)
    np.testing.assert_equal(acc.class_prediction(slice_scores), predicted_classes)

    slice_scores = np.arange(5.5, 9, 0.5)
    predicted_classes = np.array([3.0, 4.0, 4.0, 4.0, 4.0, np.nan, np.nan])
    np.testing.assert_equal(acc.class_prediction(slice_scores), predicted_classes)


def test_ground_truth_class():
    estimated_landmark_slice_scores = np.array([-5, -2, 0, 1, 4])
    landmark_positions = np.array([np.nan, np.nan, np.nan, 5, 8])
    class_to_landmark = {0: [0, 1], 1: [1, 3], 2: [3, 4]}

    acc = Accuracy(estimated_landmark_slice_scores, class_to_landmark)
    ground_truth_classes = acc.ground_truth_class(landmark_positions, max_slices=10)
    np.testing.assert_equal(
        ground_truth_classes,
        np.array(
            [np.nan, np.nan, np.nan, np.nan, np.nan, 2.0, 2.0, 2.0, np.nan, np.nan]
        ),
    )


def test_volume():
    class_to_landmark = {0: [0, 1], 1: [1, 3], 2: [3, 4]}
    expected_scores = np.array([-5, -2.5, 0, 1, 3])
    acc = Accuracy(expected_scores, class_to_landmark)

    slice_scores = np.array([-7, -6.5, -6, -5, -4, -3, -2, -1, -0.5, 0, 1, 1.2, 1.5, 2])
    landmarks = np.array([3, 5, 6, 8, np.nan])
    class_prediction = np.array(
        [np.nan, np.nan, np.nan, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2]
    )
    groundtruth_prediction = np.array(
        [
            np.nan,
            np.nan,
            np.nan,
            0,
            0,
            1,
            1,
            1,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        ]
    )
    np.testing.assert_equal(acc.class_prediction(slice_scores), class_prediction)
    np.testing.assert_equal(
        acc.ground_truth_class(landmarks, len(slice_scores)), groundtruth_prediction
    )
    assert acc.volume(slice_scores, landmarks) == 0.8



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import numpy as np
from numpy.lib.arraysetops import isin
import pandas as pd
import random, math, cv2
import albumentations as A
from torch.utils.data import Dataset

cv2.setNumThreads(1)


class BaseDataset(Dataset):
    """Parent class to create a dataset for a body part regression model.

    Args:
        data_path (str, optional): Path where .npy data is stored. Defaults to "".
        filenames (list, optional): filenames of .npy volumes. Defaults to [].
        z_spacings (list, optional): corresponding z-spacings to .npy volumes. Defaults to [].
        landmark_path (str, optional): path to landmark-file which stores the annotated landmarks for a
        subset of the training data, for the validation data and for the test data. Defaults to None.
        landmark_sheet_name (str, optional): sheet-name of landmark file which corresponds to dataset e.g. val-landmarks . Defaults to "".
        num_slices (int, optional): Numer of sampled slices per volume.
        equidistance_range (list, optional): Distance of sampling num_slices from volume.
        custom_transform (bool, optional): custom slice wise transformations. Defaults to False.
        albumentation_transform (bool, optional): slice wise transformations from albumentation. Defaults to False.
        random_seed (int, optional): define random seed for sampling. Defaults to 0.
    """

    def __init__(
        self,
        data_path: str = "",
        filenames: list = [],
        z_spacings: list = [],
        landmark_path: str = None,
        landmark_sheet_name: str = "",
        num_slices: int = 4,
        equidistance_range: list = [5, 100],
        custom_transform=False,
        albumentation_transform=False,
        random_seed: int = 0,
    ):

        self.data_path = data_path
        self.filenames = filenames
        self.filepaths = [os.path.join(data_path, f) for f in filenames]
        self.z_spacings = z_spacings  # in mm
        self.length = len(self.filepaths)
        self.num_slices = num_slices
        self.equidistance_range = equidistance_range
        self.custom_transform = custom_transform
        self.random_seed = random_seed
        random.seed(random_seed)

        # define landmark related
        if landmark_path:
            self.landmark_df = pd.read_excel(
                landmark_path,
                sheet_name=landmark_sheet_name,
                engine="openpyxl",
                index_col="filename",
            )
            self.landmark_matrix = np.array(self.landmark_df)
            self.landmark_names = self.landmark_df.columns
            self.landmark_files = [
                f + ".npy" for f in self.landmark_df.index if isinstance(f, str)
            ]
            self.landmark_ids = [
                filename_to_id(f, filenames) for f in self.landmark_files
            ]
            (
                self.landmark_slices_per_volume,
                self.defined_landmarks_per_volume,
            ) = self.get_landmark_slices()

        # define augmentations
        if custom_transform:
            self.custom_transform = custom_transform

        # Use identity function, if no custom transformation is defined
        else:
            self.custom_transform = lambda x: x

        if albumentation_transform:
            self.albumentation_transform = albumentation_transform

        # Use identity function, if no transformation is defined
        else:
            self.albumentation_transform = A.Compose([A.Transpose(p=0)])

    def __len__(self):
        return self.length

    def get_full_volume(self, idx: int):
        filepath = self.filepaths[idx]
        volume = np.load(parse2plainname(filepath) + ".npy") 
        return swap_axis(volume)

    def get_landmark_idx(self, idx: int):
        filename = self.filenames[idx]
        idx = np.where(filename == np.array(self.landmark_files))[0]
        if len(idx) == 1:
            return idx[0]
        return np.nan

    def get_landmark_slices(self):
        landmark_slices_per_volume = []
        defined_landmarks_per_volume = []

        for i, file in enumerate(self.landmark_files):
            # get slice indices of defined landmarks for file
            landmark_indices = self.landmark_matrix[i, :]

            # get information about which landmarks are defined e.g. 0, 1, 2, 3
            defined_landmarks = np.where(~np.isnan(landmark_indices))[0]

            # only save indices of defined landmarks to drop missing values and convert to int
            landmark_indices = landmark_indices[defined_landmarks].astype(int)

            # extract slices of landmark positions for file
            slices = get_slices(os.path.join(self.data_path, file), landmark_indices)

            landmark_slices_per_volume.append(slices)
            defined_landmarks_per_volume.append(defined_landmarks)

        return landmark_slices_per_volume, defined_landmarks_per_volume


def get_full_volume_from_filepath(filepath: str):
    volume = np.load(filepath)
    return swap_axis(volume)


def get_slices(filepath, indices):
    volume = np.load(parse2plainname(filepath) + ".npy", mmap_mode="r")
    x = volume[:, :, indices]
    return swap_axis(x)


def swap_axis(x):
    return x.swapaxes(2, 1).swapaxes(1, 0)


def filename_to_id(filename, filename_array):
    filename = parse2plainname(filename)
    filename_array = parse2plainname(filename_array)


    ids = np.where(filename == np.array(filename_array))[0]
    if len(ids) == 0:
        raise ValueError(f"filename {filename} is not in the filename list: {filename_array}")
    else:
        return ids[0]


def parse2plainname(value): 
    if isinstance(value, str): 
        value = value.replace(".nii", "").replace(".gz", "").replace(".npy", "")
    else: 
        value = [v.replace(".nii", "").replace(".gz", "").replace(".npy", "") for v in value]
    return value




"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pandas as pd
import random, math, cv2
import datetime, os, sys


cv2.setNumThreads(1)

sys.path.append("../../")
from bpreg.dataset.base_dataset import BaseDataset, swap_axis


class SSBRDataset(BaseDataset):
    def __init__(
        self,
        data_path: str = "",
        filenames: list = [],
        z_spacings: list = [],
        landmark_path: str = "",
        landmark_sheet_name: str = "",
        num_slices: int = 8,
        equidistance_range: list = [2, 10],
        custom_transform=False,
        albumentation_transform=False,
        random_seed: int = 0,
    ):

        BaseDataset.__init__(
            self,
            data_path,
            filenames,
            z_spacings,
            landmark_path,
            landmark_sheet_name,
            num_slices,
            equidistance_range,
            custom_transform,
            albumentation_transform,
            random_seed,
        )

    def __getitem__(self, idx):
        filepath = self.filepaths[idx]
        z_spacing = self.z_spacings[idx]
        volume = np.load(filepath, mmap_mode="r")
        indices = self.get_random_slice_indices(volume.shape[2])
        x = volume[:, :, indices]

        # transform each slice seperatly
        for i in range(x.shape[2]):
            x[:, :, i] = self.custom_transform(x[:, :, i])
            x[:, :, i] = self.albumentation_transform(image=x[:, :, i])["image"]

        return swap_axis(x), indices, np.nan

    def get_random_slice_indices(self, z_slices):
        # choose distance k between slices
        max_sampled_slices_for_volume = z_slices // self.num_slices
        k = random.randint(
            self.equidistance_range[0],
            min(self.equidistance_range[1], max_sampled_slices_for_volume),
        )

        slice_range = self.num_slices * k
        min_starting_slice = 0
        max_starting_slice = z_slices - slice_range

        # randomly choose the start slice
        starting_slice = random.randint(min_starting_slice, max_starting_slice)

        slice_indices = np.arange(
            starting_slice, starting_slice + self.num_slices * k, k
        )
        return slice_indices



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np


class LMSE:
    """class to calculate the landmark mean square error.
    The landmark mean square error is a normalized verison of the mean square error.
    As normalization constant d the difference between the expected slice scores of two landmarks is used.
    """

    def __init__(self):
        pass

    def from_dataset(self, model, val_dataset, train_dataset):
        val_score_matrix = model.compute_slice_score_matrix(val_dataset)
        train_score_matrix = model.compute_slice_score_matrix(train_dataset)
        d = self.get_normalizing_constant(np.nanmean(train_score_matrix, axis=0))
        mse, mse_std = self.from_matrices(val_score_matrix, train_score_matrix)

        return mse, mse_std, d

    def lmse_per_landmark_from_matrices(self, score_matrix, reference_matrix, d=False):
        square_error_matrix = self.get_square_error_matrix(
            score_matrix, reference_matrix, d=d
        )
        lmse_per_landmark = np.nanmean(square_error_matrix, axis=0)

        counts = np.sum(np.where(~np.isnan(square_error_matrix), 1, 0), axis=0)
        lmse_errors = np.nanstd(square_error_matrix, ddof=1, axis=0) / np.sqrt(counts)

        return lmse_per_landmark, lmse_errors

    def lmse_per_volume_from_matrices(self, score_matrix, reference_matrix, d=False):
        square_error_matrix = self.get_square_error_matrix(
            score_matrix, reference_matrix, d=d
        )
        lmse_per_volume = np.nanmean(square_error_matrix, axis=1)
        counts = np.sum(np.where(~np.isnan(square_error_matrix), 1, 0), axis=1)
        lmse_errors = np.nanstd(square_error_matrix, ddof=1, axis=1) / np.sqrt(counts)

        return lmse_per_volume, lmse_errors

    def lmse_per_slice_from_matrices(self, score_matrix, reference_matrix, d=False):
        square_error_matrix = self.get_square_error_matrix(
            score_matrix, reference_matrix, d=d
        )
        lmse = np.nanmean(square_error_matrix)
        counts = np.sum(np.where(~np.isnan(square_error_matrix), 1, 0))
        lmse_std = np.nanstd(square_error_matrix, ddof=1) / np.sqrt(counts)

        return lmse, lmse_std

    def from_matrices(self, score_matrix, reference_matrix, d=False):
        lmses, lmse_stds = self.lmse_per_volume_from_matrices(
            score_matrix, reference_matrix, d=d
        )
        return np.mean(lmses), np.std(lmses, ddof=1) / np.sqrt(len(lmses))

    def get_square_error_matrix(self, score_matrix, reference_matrix, d=False):
        expected_scores = np.nanmean(reference_matrix, axis=0)
        if not d:
            d = np.abs(self.get_normalizing_constant(expected_scores))
        square_error_matrix = self.from_instance(expected_scores, score_matrix, d)
        return square_error_matrix

    def from_volume(self, landmarks, scores, expected_scores):
        d = self.get_normalizing_constant(expected_scores)
        expected_scores = expected_scores[~np.isnan(landmarks)]
        landmarks = np.array(landmarks[~np.isnan(landmarks)]).astype(int)
        scores = scores[landmarks]
        return self.from_instance(scores, expected_scores, d)

    def from_instance(self, y_truth, y_pred, d):
        mse = ((y_truth - y_pred) / d) ** 2
        return mse

    def get_normalizing_constant(self, expected_scores):
        return (expected_scores[-1] - expected_scores[0]) / 100




import numpy as np
from bpreg.score_processing import Scores

def create_scores(slope, error, n, zspacing, b=0): 
    epsilon = np.random.normal(0, error, n) 
    x = np.arange(0, n, 1) 
    y = x*zspacing*slope + epsilon + b
    
    return y 

def test_basic_score_processing(): 
    slope=0.11
    zspacing=1
    y = create_scores(slope=slope, error=0.2, n=100, zspacing=zspacing)
    
    scores = Scores(y, zspacing=zspacing, transform_min=0, transform_max=100, slope_mean=slope)

    # Check if no missing values are in cleaned slice scores 
    assert np.sum(np.isnan(scores.values)*1) == 0

    # check if zspacing is valid 
    assert scores.valid_zspacing == 1

def test_tail_missing_values(): 
    zspacing=4
    slope=0.11
    y = create_scores(slope=0.11, error=0.2, n=400, zspacing=zspacing, b=-20)
    y = [20, 20, 20] + list(y) + [20, 20, 20] 
    
    scores = Scores(y, zspacing=zspacing, transform_min=0, transform_max=100, slope_mean=slope)
    assert np.sum(np.isnan(scores.values[0:3])) == 3
    assert np.sum(np.isnan(scores.values[-3:])) == 3

def test_transform(): 
    y = np.arange(0, 10)
    scores = Scores(y, zspacing=1, transform_min=1, transform_max=8, slope_mean=1)
    assert scores.original_transformed_values[1] == 0
    assert scores.original_transformed_values[-2] == 100

def test_valid_zspacing(): 
    slope=0.11
    zspacing=1
    y = create_scores(slope=slope, error=0.2, n=100, zspacing=zspacing)
    
    scores = Scores(y, zspacing=zspacing, transform_min=0, transform_max=100, slope_mean=2*slope)
    assert scores.valid_zspacing==0
    
    scores = Scores(y, zspacing=zspacing, transform_min=0, transform_max=100, slope_mean=0.5*slope)
    assert scores.valid_zspacing==0




import numpy as np
import cv2, os, json

cv2.setNumThreads(1)

abspath = os.path.abspath(__file__)
MAIN_PATH = os.path.dirname(os.path.dirname(os.path.dirname(abspath)))

# Path for default inference model
DEFAULT_MODEL = os.path.join(MAIN_PATH, "src/models/public_bpr_model/")

# define landmark names
LANDMARK_NAMES = [
    "pelvis-start",
    "femur-end",
    "L5",
    "L3",
    "L1",
    "Th11",
    "Th8",
    "Th5",
    "Th2",
    "C6",
    "C1",
    "eyes-end",
]

# define classes
# 0 - legs, 1 - pelvis, 2 - abdomen
CLASSES = {"legs": 0, "pelvis": 1, "abdomen": 2, "thorax": 3, "neck": 4, "head": 5}

LANDMARK_CLASS_MAPPING = {
    0: CLASSES["legs"],
    1: CLASSES["pelvis"],
    2: CLASSES["pelvis"],
    3: CLASSES["abdomen"],
    4: CLASSES["abdomen"],
    5: CLASSES["abdomen"],
    6: CLASSES["thorax"],
    7: CLASSES["thorax"],
    8: CLASSES["thorax"],
    9: CLASSES["neck"],
    10: CLASSES["neck"],
    11: CLASSES["head"],
}

CLASS_TO_LANDMARK_5 = {
    CLASSES["pelvis"]: [0, 2],
    CLASSES["abdomen"]: [2, 5],
    CLASSES["thorax"]: [5, 8],
    CLASSES["neck"]: [8, 10],
    CLASSES["head"]: [10, 11],
}

CLASS_TO_LANDMARK_3 = {
    CLASSES["pelvis"]: [0, 2],
    CLASSES["abdomen"]: [2, 5],
    CLASSES["thorax"]: [5, 8],
}

# define colors for plots
COLORS = [
    "black",
    "gray",
    "silver",
    "lightcoral",
    "brown",
    "chocolate",
    "goldenrod",
    "red",
    "lightgreen",
    "green",
    "deepskyblue",
    "steelblue",
    "mediumorchid",
    "plum",
    "purple",
    "pink",
]

DF_DATA_SOURCE_PATH = "/home/AD/s429r/Documents/Data/DataSet/MetaData/meta-data-public-dataset-npy-arrays-3.5mm-windowing-sigma.xlsx"
LANDMARK_PATH = (
    "/home/AD/s429r/Documents/Data/DataSet/MetaData/landmarks-meta-data-v2.xlsx"
)
DATA_PATH = "/home/AD/s429r/Documents/Data/DataSet/Arrays-3.5mm-sigma-01/"

BODY_PARTS = {
    "legs": [np.nan, "pelvis_start"],
    "pelvis": ["pelvis_start", "pelvis_end"],
    "abdomen": ["L5", "Th8"],
    "chest": ["Th12", "Th1"],
    "shoulder-neck": ["Th3", "C2"],
    "head": ["C5", np.nan],
}

TRANSFORM_STANDARD_PARAMS = {
    "GaussNoise": {
        "std_min": 0,
        "std_max": 0.04,
        "min_value": -1,
        "max_value": 1,
        "p": 0.5,
    },
    "ShiftHU": {"limit": 0.08, "p": 0.5, "min_value": -1, "max_value": 1},
    "ScaleHU": {"p": 0.5, "scale_delta": 0.2, "max_value": 1, "min_value": -1},
    "AddFrame": {"p": 0.25, "dimension": 128, "r_circle": 0.75, "fill_value": -1},
    "Flip": {"p": 0.5},
    "Transpose": {"p": 0.5},
    "ShiftScaleRotate": {
        "shift_limit": 0,
        "scale_limit": 0.2,
        "rotate_limit": 10,
        "p": 0.5,
        "border_mode": cv2.BORDER_REFLECT_101,
    },
    "GaussianBlur": {
        "blur_limit": (3, 7),
        "sigma_limit": 0.5,
        "always_apply": False,
        "p": 0.5,
    },
}



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys, os
import numpy as np

sys.path.append("../../")
from bpreg.settings.settings import *


class BodyPartExaminedDict:
    def __init__(self, lookuptable: dict, body_parts=BODY_PARTS):
        self.lookuptable = lookuptable
        self.landmarkDict = body_parts
        self.scoreDict = get_scoreDict(self.landmarkDict, self.lookuptable)

    def get_score_indices(self, scores, min_score=np.nan, max_score=np.nan):
        scores = np.array(scores)
        if ~np.isnan(min_score) & ~np.isnan(max_score):
            return np.where((scores >= min_score) & (scores < max_score))
        elif np.isnan(min_score) & ~np.isnan(max_score):
            return np.where(scores < max_score)
        elif ~np.isnan(min_score) & np.isnan(max_score):
            return np.where(scores > min_score)
        else:
            return np.arange(0, len(scores))

    def get_examined_body_part(self, scores):
        bodyPartDict = {}
        for bodypart, boundary_scores in self.scoreDict.items():
            indices = self.get_score_indices(
                scores, min_score=boundary_scores[0], max_score=boundary_scores[1]
            )[0]
            bodyPartDict[bodypart] = list(indices.astype(np.float64))

        return bodyPartDict


def get_scoreDict(landmarkDict: dict, lookuptable: dict) -> dict:
    scoreDict = {}
    for key, items in landmarkDict.items():
        scores = []
        for landmark in items:
            if isinstance(landmark, float):
                scores.append(landmark)
            else:
                scores.append(lookuptable[landmark]["mean"])

        scoreDict[key] = scores
    return scoreDict



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pandas as pd
import pydicom
import argparse
import glob
import os
import SimpleITK as sitk


class DicomHandler:
    def __init__(self):
        pass

    def base_df(self, path, study, suffix=""):
        df = self.dir2df(path)
        df = df.drop_duplicates("base_path").reset_index(drop=True)
        df.index = df["base_path"]

        df["dirname"] = df["base_path"].apply(
            lambda x: suffix + x.replace(path, "").replace("/", "__").replace(" ", "--")
        )
        df["broken_path"] = 0
        df["id_name"] = [study + "_" + i for i in np.arange(len(df)).astype(str)]
        df["success"] = 0
        df.drop(["file_path", "file_name"], axis=1, inplace=True)

        return df

    def metaData2Series(self, path):
        """
        Input: pydicom Dataset
        Output: Pandas Series
        """

        def func(x):
            if isinstance(x, pydicom.multival.MultiValue):
                x = str(x)
            return x

        ds = pydicom.dcmread(path)
        dcmDict = {key: func(ds[key].value) for key in ds.dir() if key != "PixelData"}
        series = pd.Series(dcmDict)

        return series

    def dir2df(self, path):
        file_paths = glob.glob(path + "/**/*.dcm", recursive=True)
        file_names = [f.split("/")[-1] for f in file_paths]
        base_paths = [
            file_paths[i].replace(file_names[i], "") for i in range(len(file_paths))
        ]

        df = pd.DataFrame(index=range(0, len(file_paths)))
        df["file_path"] = file_paths
        df["file_name"] = file_names
        df["base_path"] = base_paths
        df["images_in_dir"] = df["base_path"].map(
            df.groupby("base_path")["file_name"].count()
        )

        return df

    def add_unknown_columns(self, df, series):
        colsNotInDf = list(set(series.keys()) - set(df.columns))
        for col in colsNotInDf:
            df[col] = np.nan

    def dicom2nifti(self, ifilepath, ofilepath):
        reader = sitk.ImageSeriesReader()
        dicom_names = reader.GetGDCMSeriesFileNames(ifilepath)
        reader.SetFileNames(dicom_names)
        image = reader.Execute()
        sitk.WriteImage(image, ofilepath)

    def dicom2array(self, ifilepath):
        reader = sitk.ImageSeriesReader()
        dicom_names = reader.GetGDCMSeriesFileNames(ifilepath)
        reader.SetFileNames(dicom_names)
        image = reader.Execute()
        array = sitk.GetArrayFromImage(image)
        return array

    def nrrd2nifti(self, ifilepath, ofilepath):
        reader = sitk.ImageFileReader()
        reader.SetImageIO("NrrdImageIO")
        reader.SetFileName(ifilepath)
        image = reader.Execute()
        sitk.WriteImage(image, ofilepath)

    def get_first_image(self, x, case="default"):

        if case == "CQ500":
            return "CT000001.dcm"

        filename = "1-1.dcm"
        for i, y in enumerate([10, 100, 1000, 10000], 1):
            if x / y >= 1:
                filename = "1-" + "0" * i + "1.dcm"

        return filename

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", default="")
    parser.add_argument("-o", default="")

    value = parser.parse_args()
    input_dirpath = value.i
    output_dirpath = value.o

    dh = DicomHandler()
    for file in os.listdir(input_dirpath): 
        print(file)
        ifilepath = os.path.join(input_dirpath, file)
        ofilepath = os.path.join(output_dirpath, file.replace(".dcm", ".nii.gz"))
        dh.dicom2nifti(ifilepath, ofilepath)



from setuptools import setup, find_packages

# read readme
with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="bpreg",
    version="0.0.6", 
    packages=find_packages(),
    url="https://github.com/MIC-DKFZ/BodyPartRegression",
    include_package_data=True,
    package_data={'bpreg': ['settings/body-part-metadata.md']}, 
    test_suite="unittest",
    install_requires=[
        "pytorch_lightning==1.2.10", 
        "nibabel==3.2.1", 
        "scipy==1.7.0", 
        "albumentations==0.5.2", 
        "dataclasses", 
        "pandas==1.2.1", 
        "torch==1.8.1",  
        "torchvision==0.9.1",
    ],
    data_files=[('models', ['src/models/README.md', 'bpreg/settings/body-part-metadata.md'])], 
    long_description=readme,
    long_description_content_type = "text/markdown",
    author="Division of Medical Image Computing, German Cancer Research Center",
    author_email="mic-office@dkfz.de", 
    maintainer_email="s.schuhegger@dkfz-heidelberg.de",
    entry_points={
        "console_scripts": [
            "bpreg_predict = bpreg.scripts.bpreg_inference:main",
        ]
    },
)



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np


def linear_transform(x, scale=1, min_value=0, max_value=1):
    x = x - min_value
    x = x * scale / (max_value - min_value)

    return x


def transform_0to100(
    score,
    lookuptable,
    min_value=np.nan,
    landmark_start="pelvis_start",
    landmark_end="eyes_end",
):
    if np.isnan(min_value):
        min_value = lookuptable[landmark_start]["mean"]
    max_value = lookuptable[landmark_end]["mean"]

    score = linear_transform(score, scale=100, min_value=min_value, max_value=max_value)
    return score


def transform_lookuptable(
    lookuptable, landmark_start="pelvis_start", landmark_end="eyes_end"
):
    lookup_copy = {key: {} for key in lookuptable}
    for key in lookuptable:
        lookup_copy[key]["mean"] = np.round(
            transform_0to100(
                lookuptable[key]["mean"],
                lookuptable,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            3,
        )
        lookup_copy[key]["std"] = np.round(
            transform_0to100(
                lookuptable[key]["std"],
                lookuptable,
                min_value=0.0,
                landmark_start=landmark_start,
                landmark_end=landmark_end,
            ),
            3,
        )

    return lookup_copy



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os, sys
import argparse
import torch
from shutil import copyfile

sys.path.append("../")
from bpreg.inference.inference_model import InferenceModel
from bpreg.settings import *
from bpreg.evaluation.visualization import plot_scores

def bpreg_for_directory(
    model_path: str,
    input_dirpath: str,
    output_dirpath: str,
    skip_existing: bool = True,
    stringify_json: bool = False,
    gpu_available: bool=True, 
):
    # test if gpu is available
    if not torch.cuda.is_available(): gpu_available = False 
    model = InferenceModel(model_path, gpu=gpu_available)

    if input_dirpath == "": 
        raise ValueError("Input path is not defined. Please define the input path via -i <input_path>.")

    if output_dirpath == "": 
        raise ValueError("Output path is not defined. Please define the output path via -o <output_path>.")

    ifiles = [f for f in os.listdir(input_dirpath) if f.endswith((".nii.gz", ".nii"))]
    ofiles = [f.replace(".nii", "").replace(".gz", "") + ".json" for f in ifiles]

    for ifile, ofile in zip(ifiles, ofiles):
        ipath = os.path.join(input_dirpath, ifile)
        opath = os.path.join(output_dirpath, ofile)

        if os.path.exists(opath) and skip_existing == 1:
            print(f"JSON-file already exists. Skip file: {ifile}")
            continue

        print(f"Create body-part meta data file: {ofile}")
        model.nifti2json(ipath, opath, stringify_json=stringify_json)



def plot_scores_in_json_files(output_path): 
    json_files = [f for f in os.listdir(output_path) if f.endswith(".json")]
    for file in json_files: 
        save_path = os.path.join(output_path, file.replace(".json", ".png"))
        plot_scores(os.path.join(output_path, file), save_path=save_path)

def bpreg_inference(
    input_path: str,
    output_path: str,
    model: str=DEFAULT_MODEL,
    skip_existing: bool=True,
    stringify_json: bool=False,
    gpu_available: bool=True, 
    plot: bool=False, 
):

    # run body part regression for each file in the dictionary
    bpreg_for_directory(
        model,
        input_path,
        output_path,
        skip_existing=skip_existing,
        stringify_json=stringify_json,
        gpu_available=gpu_available
    )

    # plot slice scores if plot is True
    if plot: 
        print("Plot slice scores. ")
        plot_scores_in_json_files(output_path)

    # copy documentation for json metadata files into repository 
    copyfile(os.path.join(MAIN_PATH, "bpreg/settings/body-part-metadata.md"), os.path.join(output_path, "README.md"))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", default=DEFAULT_MODEL)
    parser.add_argument("--plot", default=False)
    parser.add_argument("-i", default="")
    parser.add_argument("-o", default="")
    parser.add_argument("--skip", default=True)
    parser.add_argument("--str", default=False)
    parser.add_argument("--gpu", default=True)

    value = parser.parse_args()
    model_path = value.model
    plot = value.plot
    input_dirpath = value.i
    output_dirpath = value.o
    skip_existing = value.skip
    stringify_json = value.str
    gpu_available = value.gpu

    bpreg_inference(
        input_dirpath, output_dirpath, model_path, skip_existing, stringify_json, gpu_available=gpu_available, plot=plot
    )


if __name__ == "__main__":
    main()



"""
Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0
   
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
from bpreg.inference.inference_settings import postprocess_model_for_inference 
from bpreg.settings import * 

def main():
   """
   create inference-settings.json file for model to use model for inference. 
   """ 
   parser = argparse.ArgumentParser()
   parser.add_argument("--model", default=DEFAULT_MODEL)
   parser.add_argument("--data", default=DF_DATA_SOURCE_PATH)
   parser.add_argument("--landmarks", default=LANDMARK_PATH)
   parser.add_argument("--uq", default=0.995)
   parser.add_argument("--lq", default=0.005)
   parser.add_argument("--minl", default="pelvis_start")
   parser.add_argument("--maxl", default="eyes_end")

   value = parser.parse_args()
   model_path = value.model
   df_data_source_path = value.data
   landmark_path = value.landmarks
   upper_quantile = value.uq
   lower_quantile = value.lq
   min_landmark = value.minl
   max_landmark = value.maxl

   postprocess_model_for_inference(model_path, 
                                 df_data_source_path, 
                                 landmark_path, 
                                 upper_tangential_slope_quantile=upper_quantile, 
                                 lower_tangential_slope_quantile=lower_quantile, 
                                 transform_min_landmark=min_landmark, 
                                 transform_max_landmark=max_landmark, 
                                 )
